/******************************************************************************\
|*                                                                            *|
|*                         Neurotechnologija Core 2.0                         *|
|*                                                                            *|
|* NTypes.h                                                                   *|
|* Types and various defines definition                                       *|
|*                                                                            *|
|* Copyright (C) 2005-2006 Neurotechnologija                                  *|
|*                                                                            *|
\******************************************************************************/

#ifndef N_TYPES_H_INCLUDED
#define N_TYPES_H_INCLUDED

// C++ detection
#ifdef __cplusplus
#define N_CPP
#endif

#ifdef N_CPP
extern "C"
{
#endif

// Windows detection
#ifdef _WIN32
#define N_WINDOWS
#endif

// Mac detection
#ifdef __APPLE__
#define N_MAC
#endif

// Compiler detection

#if (_MSC_VER >= 1300)
#define N_MSVC
#endif

#ifdef __GNUC__
#define N_GCC
#endif

#if defined(__STDC__) && !defined(N_GCC)
#define N_ANSI_C
#endif

// Processor detection

#ifdef __POWERPC__
#define N_BIG_ENDIAN
#endif

#ifdef _M_X64
#define N_64
#endif

//#define N_NO_FLOAT
//#define N_NO_INT_64

#if defined(N_64) && defined(N_NO_INT_64)
#error N_64 and N_NO_INT_64 defined simultaneously
#endif

// Integer types

#ifdef N_MSVC

typedef unsigned __int8  NUInt8;
typedef signed   __int8  NInt8;
typedef unsigned __int16 NUInt16;
typedef signed   __int16 NInt16;
typedef unsigned __int32 NUInt32;
typedef signed   __int32 NInt32;

#define N_UINT8_MIN 0x00ui8
#define N_UINT8_MAX 0xFFui8
#define N_INT8_MIN 0x80i8
#define N_INT8_MAX 0x7Fi8
#define N_UINT16_MIN 0x0000ui16
#define N_UINT16_MAX 0xFFFFui16
#define N_INT16_MIN 0x8000i16
#define N_INT16_MAX 0x7FFFi16
#define N_UINT32_MIN 0x00000000ui32
#define N_UINT32_MAX 0xFFFFFFFFui32
#define N_INT32_MIN 0x80000000i32
#define N_INT32_MAX 0x7FFFFFFFi32

#else

typedef unsigned char  NUInt8;
typedef signed   char  NInt8;
typedef unsigned short NUInt16;
typedef signed   short NInt16;
typedef unsigned int   NUInt32;
typedef signed   int   NInt32;

#define N_UINT8_MIN ((NUInt8)0x00u)
#define N_UINT8_MAX ((NUInt8)0xFFu)
#define N_INT8_MIN ((NInt8)0x80)
#define N_INT8_MAX ((NInt8)0x7F)
#define N_UINT16_MIN ((NUInt16)0x0000u)
#define N_UINT16_MAX ((NUInt16)0xFFFFu)
#define N_INT16_MIN ((NInt16)0x8000)
#define N_INT16_MAX ((NInt16)0x7FFF)
#define N_UINT32_MIN 0x00000000u
#define N_UINT32_MAX 0xFFFFFFFFu
#define N_INT32_MIN 0x80000000
#define N_INT32_MAX 0x7FFFFFFF

#endif

#ifndef N_NO_INT_64

#ifdef N_MSVC

typedef unsigned __int64 NUInt64;
typedef signed   __int64 NInt64;

#define N_UINT64_MIN 0x0000000000000000ui64
#define N_UINT64_MAX 0xFFFFFFFFFFFFFFFFui64
#define N_INT64_MIN 0x8000000000000000i64
#define N_INT64_MAX 0x7FFFFFFFFFFFFFFFi64

#else

typedef unsigned long long NUInt64;
typedef signed   long long NInt64;

#define N_UINT64_MIN 0x0000000000000000ull
#define N_UINT64_MAX 0xFFFFFFFFFFFFFFFFull
#define N_INT64_MIN 0x8000000000000000ll
#define N_INT64_MAX 0x7FFFFFFFFFFFFFFFll

#endif

#endif

// Shorthand integer types

typedef NUInt8 NByte;
typedef NInt8 NSByte;
typedef NUInt16 NUShort;
typedef NInt16 NShort;
typedef NUInt32 NUInt;
typedef NInt32 NInt;

#ifndef N_NO_INT_64

typedef NUInt64 NULong;
typedef NInt64 NLong;

#endif

#define N_BYTE_MIN N_UINT8_MIN
#define N_BYTE_MAX N_UINT8_MAX
#define N_SBYTE_MIN N_INT8_MIN
#define N_SBYTE_MAX N_INT8_MAX
#define N_USHORT_MIN N_UINT16_MIN
#define N_USHORT_MAX N_UINT16_MAX
#define N_SHORT_MIN N_INT16_MIN
#define N_SHORT_MAX N_INT16_MAX
#define N_UINT_MIN N_UINT32_MIN
#define N_UINT_MAX N_UINT32_MAX
#define N_INT_MIN N_INT32_MIN
#define N_INT_MAX N_INT32_MAX

#ifndef N_NO_INT_64

#define N_ULONG_MIN N_UINT64_MIN
#define N_ULONG_MAX N_UINT64_MAX
#define N_LONG_MIN N_INT64_MIN
#define N_LONG_MAX N_INT64_MAX

#endif

#ifndef N_NO_FLOAT

// Float types
typedef float NSingle;
typedef double NDouble;

#define N_SINGLE_MIN 1.175494351e-38F
#define N_SINGLE_MAX 3.402823466e+38F
#define N_SINGLW_EPSILON 1.192092896e-07F
#define N_DOUBLE_MIN 2.2250738585072014e-308
#define N_DOUBLE_MAX 1.7976931348623158e+308
#define N_DOUBLE_EPSILON 2.2204460492503131e-016

// Shorthand float type
typedef NSingle NFloat;

#define N_FLOAT_MIN N_SINGLE_MIN
#define N_FLOAT_MAX N_SINGLE_MAX
#define N_FLOAT_EPSILON N_SINGLE_EPSILON

#endif

// Boolean types
typedef int NBoolean;

// Boolean constants
#define NTrue 1
#define NFalse 0

// Shorthand bool type
typedef NBoolean NBool;

// Character types
typedef char NAChar;

// Char type
typedef NAChar NChar;

// Size type
#ifdef N_64

typedef NUInt64 NSizeType;

#define N_SIZE_TYPE_MIN N_UINT64_MIN
#define N_SIZE_TYPE_MAX N_UINT64_MAX

#else

#ifdef N_MSVC
typedef __w64 NUInt32 NSizeType;
#else
typedef NUInt32 NSizeType;
#endif

#define N_SIZE_TYPE_MIN N_UINT32_MIN
#define N_SIZE_TYPE_MAX N_UINT32_MAX

#endif

// Pos type
#ifndef N_NO_INT_64

typedef NInt64 NPosType;

#define N_POS_TYPE_MIN N_INT64_MIN
#define N_POS_TYPE_MAX N_INT64_MAX

#else

#ifdef N_MSVC
typedef __w64 NInt32 NPosType;
#else
typedef NInt32 NPosType;
#endif

#define N_POS_TYPE_MIN N_INT32_MIN
#define N_POS_TYPE_MAX N_INT32_MAX

#endif

// NULL
#ifndef NULL
	#define NULL 0
#endif

// Result type
typedef NInt NResult;

// API and callback functions' calling convention

#ifdef N_WINDOWS

#define N_API __stdcall
#define N_CALLBACK __stdcall

#else

#define N_API
#define N_CALLBACK

#endif

// Handle

typedef void * NHandle;

#define N_DECLARE_HANDLE(name) typedef struct name##_ { int unused; } * name;

// Rational

typedef struct NURational_
{
	NUInt Numerator;
	NUInt Denominator;
} NURational;

typedef struct NRational_
{
	NInt Numerator;
	NInt Denominator;
} NRational;

// File access

typedef enum NFileAccess_ // Flags
{
	nfaRead = 1,
	nfaWrite = 2,
	nfaReadWrite = nfaRead | nfaWrite
} NFileAccess;

// Byte order

typedef enum NByteOrder_
{
	nboLittleEndian = 0,
	nboBigEndian = 1,
#ifdef N_BIG_ENDIAN
	nboSystem = nboBigEndian
#else
	nboSystem = nboLittleEndian
#endif

} NByteOrder;

#define NIsReverseByteOrder(byteOrder) ((byteOrder) != nboSystem)

// Index pair

typedef struct NIndexPair_
{
	NInt Index1;
	NInt Index2;
} NIndexPair;

#ifdef N_CPP
}
#endif

#endif // !N_TYPES_H_INCLUDED
