/* $Id: DBController.m 9016 2006-09-15 16:07:33Z mindaugas $ */

#include <stdlib.h>
#include <stdio.h>
#import "DBController.h"

struct list {
	struct list * next;
	NSString * name;
	int size;
	void * data;
};

@implementation DBController

- (int) initDB
{
	int i;
	currentDBItem = NULL;
	dBSize = 0;
	for (i = 0; i < 256; i++)
	{
		root[i] = NULL;
	}
	currentDBItemPos = 0;
	currentDBItemG = 0;
}

- (int) add:(NSString *)name data:(unsigned char *)d size:(int)s G:(int)g
{
	struct list * temp;
	
	temp = (struct list *)malloc(sizeof(struct list));
	if (!temp)
	{
		return -1;
	}
	
	temp->size = s;
	if (d)
	{
		temp->data = malloc(s);
		if (!temp->data)
		{
			free(temp);
			return -1;
		}
		memcpy(temp->data, d, s);
	}
	temp->name = [[NSString alloc] initWithString:name];
	temp->next = root[g];
	root[g] = temp;
	
	currentDBItem = NULL;
	
	dBSize++;
	return 0;
}

- (int) goToFirst
{
	currentDBItem = root[0];
	currentDBItemPos = 0;
	currentDBItemG = 0;
	if (!currentDBItem)
	{
		return [self goToNext];
	}
	else
	{
		return 0;
	}
}

- (int) goToFirstG:(int)G
{
	currentDBItem = root[G];
	currentDBItemPos = G;
	currentDBItemG = G;

	if (!currentDBItem)
	{
		return [self goToNextG];
	}
	else
	{
		return 0;
	}
}

- (int) goToNext
{
	if (currentDBItem && currentDBItem->next)
	{
		currentDBItem = currentDBItem->next;
		return 0;
	}
	else
	{
		currentDBItem = NULL;
	}

	while (!currentDBItem)
	{
		if (currentDBItemPos == 255)
		{
			return -1;
		}
		
		currentDBItemPos++;
		currentDBItem = root[currentDBItemPos];
	}
	
	return 0;
}

- (int) goToNextG
{
	int i;

	if (currentDBItem && currentDBItem->next)
	{
		currentDBItem = currentDBItem->next;
		return 0;
	}
	else
	{
		currentDBItem = NULL;
	}
	
	while (!currentDBItem)
	{
		if (currentDBItemG < currentDBItemPos)
		{
			i = (int)currentDBItemG - (int)currentDBItemPos + (int)currentDBItemG;
		}
		else
		{
			i = (int)currentDBItemG - (int)currentDBItemPos + (int)currentDBItemG + 1;
		}
		if (i < 0)
		{
			if (currentDBItemPos == 255)
			{
				return -1;
			}
			else
			{
				i = (int)currentDBItemPos + 1;
			}
		}
		if (i > 255)
		{
			if (currentDBItemPos == 0)
			{
				return -1;
			}
			else
			{
				i = (int)currentDBItemPos - 1;
			}
		}
		
		currentDBItemPos = i;
		currentDBItem = root[currentDBItemPos];
	}
	
	return 0;
}

- (NSString *)getCurrentDBItemName
{
	if (currentDBItem)
	{
		return currentDBItem->name;
	}
	else
	{
		return NULL;
	}
}

- (unsigned char *)getCurrentDBItemData
{
	if (currentDBItem)
	{
		return currentDBItem->data;
	}
	else
	{
		return NULL;
	}
}

- (int) getCurrentDBItemDataSize
{
	if (currentDBItem)
	{
		return currentDBItem->size;
	}
	else
	{
		return 0;
	}
}

- (int) getDBSize
{
	return dBSize;
}

- (int) clearDB
{
	struct list * temp, * tmp;
	int i;
	
	for (i = 0; i<256; i++)
	{
		if (root[i])
		{
			temp = root[i];
			while (temp)
			{
				free(temp->data);
				tmp = temp;
				[temp->name release];
				temp = temp->next;
				free(tmp);
			}
			root[i] = NULL;
		}
	}
		
	currentDBItem = NULL;
	currentDBItemPos = 0;
	currentDBItemG = 0;
	dBSize = 0;
	return 0;
}

- (int) saveToFile:(NSString *)filename
{
	int ret;
	int l, i;
	
	FILE * f;
	
	if (!(f = fopen([filename cString],"w")))
	{
		return -1;
	} 
	
	if (dBSize == 0)
	{
		return 0;
	}
	
	if ([self goToFirst] < 0)
	{
		return -2;
	}
	
	for (i = 0; i < dBSize;i++)
	{
		l = [[self getCurrentDBItemName] length];
		ret = fwrite((void *)&l, 1, sizeof(int), f);
		if (ret != sizeof(int))
		{
			fclose(f);
			return -3;
		}
		ret = fwrite((void *)[[self getCurrentDBItemName] cString], 1, l, f);
		if (ret != l)
		{
			fclose(f);
			return -4;
		}
		
		l = [self getCurrentDBItemDataSize];
		ret = fwrite((void *)&l, 1, sizeof(int), f);
		if (ret != sizeof(int))
		{
			fclose(f);
			return -5;
		}		
		
		ret = fwrite((void *)[self getCurrentDBItemData], 1, l, f);
		if (ret != l)
		{
			fclose(f);
			return -6;
		}

		if ([self goToNext] < 0)
		{
			fclose(f);
			return i;
		}
	}
	
	fclose(f);
	return 0;
}

- (int) loadFromFile:(NSString *)filename
{
	int ret;
	NResult res;
	int count;
	NSString * name;
	HNFRecord hRecord;
	NByte G;
	char * cName;
	int nL;
	unsigned char * data;
	int size;
	
	FILE * f;
	
	if (!(f = fopen([filename cString],"r")))
	{
		return -11;
	}
	
	[self clearDB];
	
	count = 0;
	
	while (1)
	{
		ret = fread((void *)&nL, 1, sizeof(int), f);
		if (feof(f))
		{
			return count;
		}
		
		if (ret != sizeof(int))
		{
			fclose(f);
			return -12;
		}
		
		cName = (char *)malloc(nL+1);
		if (!cName)
		{
			fclose(f);
			return -13;
		}
		
		memset(cName, 0, nL+1);
		
		ret = fread((void *)cName, 1, nL, f);
		if (ret != nL)
		{
			fclose(f);
			return -14;
		}
		
		name = [NSString stringWithCString:cName];
		
		ret = fread((void *)&size, 1, sizeof(int), f);
		if (ret != sizeof(int))
		{
			fclose(f);
			return -15;
		}
		
		ret = fread((void *)data, 1, size, f);
		if (ret != size)
		{
			fclose(f);
			return -16;
		}

		res = NFRecordCreateFromMemory(data, size, 0, NULL, &hRecord);
		if (res != N_OK)
		{
			fclose(f);
			return -17;
		}
		
		res = NFRecordGetG(hRecord, &G);
		if (res != N_OK)
		{
			fclose(f);
			NFRecordFree(hRecord);
			return -18;
		}

		if ([self add:name data:data size:size G:G] < 0)
		{
			fclose(f);
			return -19;
		}
		
		NFRecordFree(hRecord);
		
		free(cName);
		
		count++;
	}
	
	fclose(f);
	return 0;
}

@end
