#include "System.h"
#include "FileListEntryForm.h"
#include "VFFileList.h"

////////////////////////
//                    //
// CFileListEntryForm //
//                    //
////////////////////////

CFileListEntryForm::CFileListEntryForm()
	: mode(flefmEditEntry), many_entries(false), from_scanner(false), dif_res(false), dif_fid(false)
{
	SetMinimizeBox(false);
	SetMaximizeBox(false);
	SetFormBorder(fbDialog);
	SetFormPosition(fpOwnerCenter);

	int x = 8;
	int w = 50;
	int xx = x + w + 8;
	int y = 8;
	label_source = new CLabel();
	label_source->SetLeft(x);
	label_source->SetWidth(w);
	edit_source = new CEdit();
	edit_source->SetTop(y);
	edit_source->SetLeft(xx);
	edit_source->SetWidth(300);
	label_source->SetVCenter(edit_source->VCenter());
	y += edit_source->Height() + 8;
	
	label_resolution = new CLabel();
	label_resolution->SetText(STRING("&Resolution"));
	label_resolution->SetLeft(x);
	label_resolution->SetWidth(w);
	edit_resolution = new CEdit();
	edit_resolution->SetTop(y);
	edit_resolution->SetLeft(xx);
	edit_resolution->SetWidth(40);
	label_dpi = new CLabel();
	label_dpi->SetText(STRING("dpi"));
	label_dpi->SetLeft(edit_resolution->Right() + 4);
	y += edit_resolution->Height() + 8;
	label_resolution->SetVCenter(edit_resolution->VCenter());
	label_dpi->SetVCenter(edit_resolution->VCenter());

	label_fingerid = new CLabel();
	label_fingerid->SetText(STRING("Finger &ID"));
	label_fingerid->SetLeft(x);
	label_fingerid->SetWidth(w);
	edit_fingerid = new CEdit();
	edit_fingerid->SetTop(y);
	edit_fingerid->SetLeft(xx);
	edit_fingerid->SetWidth(200);
	y += edit_fingerid->Height() + 4;
	label_fingerid->SetVCenter(edit_fingerid->VCenter());
	label_preview = new CLabel();
	label_preview->SetLeft(xx);
	label_preview->SetTop(y);
	label_preview->SetWidth(edit_source->Width());
	y += label_preview->Height() + 12;

	button_ok = new CButton();
	button_ok->SetText(STRING("OK"));
	button_ok->SetTop(y);
	button_ok->SetDefault(true);
	button_ok->SetModalResult(IDOK);
	button_cancel = new CButton();
	button_cancel->SetText(STRING("Cancel"));
	button_cancel->SetTop(y);
	button_cancel->SetCancel(true);
	button_cancel->SetModalResult(IDCANCEL);
	button_default = new CButton();
	button_default->SetText(STRING("&Default"));
	button_default->SetTop(y);
	button_help = new CButton();
	button_help->SetText(STRING("Help"));
	button_help->SetTop(y);
	button_help->SetRight(edit_source->Right());
	button_default->SetRight(button_help->Left() - 8);
	button_cancel->SetRight(button_default->Left() - 8);
	button_ok->SetRight(button_cancel->Left() - 8);
	SetClientSize(button_help->Right() + 8, button_help->Bottom() + 8);

	AddControl(label_source);
	AddControl(edit_source);
	AddControl(label_resolution);
	AddControl(edit_resolution);
	AddControl(label_dpi);
	AddControl(label_fingerid);
	AddControl(edit_fingerid);
	AddControl(label_preview);
	AddControl(button_ok);
	AddControl(button_cancel);
	AddControl(button_default);
	AddControl(button_help);

	UpdateMode();
	UpdateFromScanner();
}

void CFileListEntryForm::UpdateMode()
{
	SetText(mode == flefmIdentify ? STRING("Identify") : mode == flefmEnroll ? STRING("Enroll")
		: (mode == flefmNewEntry ? STRING("New ") : STRING("Edit "))
		+ (many_entries ? STRING("entries") : STRING("entry")));
}

void CFileListEntryForm::UpdateSourceEnabled()
{
	edit_source->SetEnabled(!many_entries && !from_scanner);
	SetActiveControl(edit_source->Enabled() ? edit_source : edit_fingerid);
}

void CFileListEntryForm::UpdateFromScanner()
{
	label_source->SetText(from_scanner ? STRING("&Scanner") : STRING("&Filename"));
	label_resolution->SetVisible(!from_scanner);
	edit_resolution->SetVisible(!from_scanner);
	label_dpi->SetVisible(!from_scanner);
	UpdateSourceEnabled();
	UpdateSource();
}

void CFileListEntryForm::UpdateSource()
{
	string text = source;
	if(many_entries && !from_scanner)
	{
		if(!text.IsEmpty()) text += TEXT(" ");
		text += TEXT("...");
	}
	edit_source->SetText(text);
	UpdatePreview();
}

void CFileListEntryForm::UpdatePreview()
{
	try
	{
		label_preview->SetText(MakeFingerID(edit_fingerid->Text(), source, from_scanner));
	}
	catch(exception &e)
	{
		label_preview->SetText(Format(TEXT("Error: %s"), (LPCTSTR)AnsiToString(e.what())));
	}
}

bool CFileListEntryForm::OnCloseQuery()
{
	try
	{
		if(ModalResult() == IDOK)
		{
			if(edit_source->Text().IsEmpty() && !many_entries)
			{
				edit_source->SetFocus();
				throw exception("Please specify file name");
			}
			if(!from_scanner)
			{
				if(edit_resolution->Text().IsEmpty() && !(many_entries && dif_res))
				{
					edit_resolution->SetFocus();
					throw exception("Please specify resolution");
				}
				else if(!edit_resolution->Text().IsEmpty())
				{
					int i;
					try
					{
						i = StringToInt(edit_resolution->Text());
					}
					catch(exception &)
					{
						edit_resolution->SetFocus();
						throw exception("Resolution have to be an integer value");
					}
					if(i <= 0)
					{
						edit_resolution->SetFocus();
						throw exception("Resolution have to be greater than zero");
					}
				}
			}
			try
			{
				CheckFingerID(edit_fingerid->Text(), from_scanner);
			}
			catch(exception &e)
			{
				edit_fingerid->SetFocus();
				throw EXCEPTION(Format(TEXT("Error processing finger id:\n%s"), (LPCTSTR)AnsiToString(e.what())));
			}
		}
		return true;
	}
	catch(exception &e)
	{
		ErrorBox(this, e);
	}
	return false;
}

void CFileListEntryForm::OnClose()
{
	if(dif_res) dif_res = edit_resolution->Text().IsEmpty();
	if(dif_fid) dif_fid = edit_fingerid->Text().IsEmpty();
}

void CFileListEntryForm::OnHelp()
{
	MessageBox(this, GetFingerIDHelp(from_scanner), STRING("Help on finger id field"));
}

void CFileListEntryForm::OnCommand(WORD /*id*/, WORD event, CControl *control)
{
	if(control)
	{
		if(event == BN_CLICKED)
		{
			if(control == button_help) OnHelp();
			else if(control == button_default) SetDefaults();
		}
		else if(event == EN_CHANGE)
		{
			if(control == edit_fingerid || control == edit_source)
			{
				if(control == edit_source && !from_scanner && !many_entries) source = edit_source->Text();
				UpdatePreview();
			}
		}
	}
}

void CFileListEntryForm::SetDefaults()
{
	if(!from_scanner) SetResolution(VF_IMAGE_RESOLUTION);
	SetFingerID(from_scanner ? default_scanner_finger_id : default_file_finger_id);
}

void CFileListEntryForm::SetMode(FLEFMode value)
{
	if(mode != value)
	{
		mode = value;
		UpdateMode();
	}
}

void CFileListEntryForm::SetManyEntries(bool value)
{
	if(many_entries != value)
	{
		many_entries = value;
		UpdateMode();
		UpdateSourceEnabled();
	}
}

void CFileListEntryForm::SetFromScanner(bool value)
{
	if(from_scanner != value)
	{
		from_scanner = value;
		UpdateFromScanner();
	}
}

void CFileListEntryForm::SetDifRes(bool value)
{
	if(dif_res != value)
	{
		dif_res = value;
		if(dif_res) edit_resolution->SetText(emptyString);
	}
}

void CFileListEntryForm::SetDifFID(bool value)
{
	if(dif_fid != value)
	{
		dif_fid = value;
		if(dif_fid)
		{
			edit_fingerid->SetText(emptyString);
			UpdatePreview();
		}
	}
}

const string& CFileListEntryForm::Source()
{
	return source;
}

int CFileListEntryForm::Resolution()
{
	return dif_res ? -1 : StringToInt(edit_resolution->Text());
}

const string& CFileListEntryForm::FingerID()
{
	return edit_fingerid->Text();
}

void CFileListEntryForm::SetSource(const string &value)
{
	source = value;
	UpdateSource();
	UpdatePreview();
}

void CFileListEntryForm::SetResolution(int value)
{
	edit_resolution->SetText(dif_res ? emptyString : IntToString(value));
}

void CFileListEntryForm::SetFingerID(const string &value)
{
	edit_fingerid->SetText(dif_fid ? emptyString : value);
	UpdatePreview();
}

//////////////////////
//                  //
// FileListEntryBox //
//                  //
//////////////////////

bool FileListEntryBox(CForm *owner, DWORD &flags, string &source, int &resolution, string &finger_id)
{
	CFileListEntryForm *form = new CFileListEntryForm();
	form->SetMode((FLEFMode)(flags & FLEB_MODE_MASK));
	form->SetManyEntries(IsFlagSet(flags, FLEB_MANY_ENTRIES));
	form->SetFromScanner(IsFlagSet(flags, FLEB_FROM_SCANNER));
	form->SetDifRes(IsFlagSet(flags, FLEB_DIF_RES));
	form->SetDifFID(IsFlagSet(flags, FLEB_DIF_FID));
	form->SetSource(source);
	if(!form->FromScanner()) if(!form->DifRes()) form->SetResolution(resolution);
	form->SetFingerID(finger_id);
	form->SetParent(owner);
	bool result = form->ShowModal() == IDOK;
	if(result)
	{
		if(!form->ManyEntries()) source = form->Source();
		if(!form->FromScanner()) if(!form->DifRes()) resolution = form->Resolution();
		if(!form->DifFID()) finger_id = form->FingerID();
		SetFlag(flags, FLEB_DIF_RES, form->DifRes());
		SetFlag(flags, FLEB_DIF_FID, form->DifFID());
	}
	delete form;
	return result;
}
