#include "System.h"
#include "GUILog.h"

CGUILog::CGUILog()
	: CLog(), editControl(NULL)
{
}

void CGUILog::SetEditControl(CEdit * value)
{
	editControl = value;
}

void CGUILog::OnClear()
{
	CLog::OnClear();
	if(editControl) editControl->ClearLines();
}

void CGUILog::OnAddLine(const string &text)
{
	CLog::OnAddLine(text);
	if(editControl)
	{
		SCROLLINFO si;
		si.cbSize = sizeof(SCROLLINFO);
		si.fMask = SIF_ALL;
		bool needsScroll;
		if(GetScrollInfo(editControl->Handle(), SB_VERT, &si))
		{
			needsScroll = si.nPos + (int)si.nPage >= si.nMax;
		#if(WINVER >= 0x0500)
			SCROLLBARINFO sbi;
			sbi.cbSize = sizeof(SCROLLBARINFO);
			if(GetScrollBarInfo(editControl->Handle(), OBJID_VSCROLL, &sbi))
				needsScroll |= (sbi.rgstate[0] & (STATE_SYSTEM_INVISIBLE | STATE_SYSTEM_UNAVAILABLE)) != 0;
		#endif
		}
		else needsScroll = false;
		int i = editControl->LineCount() - 1;
		editControl->SetLine(i, editControl->Line(i) + text);
		editControl->AddLine(emptyString);
		if(needsScroll) editControl->SendMessage(WM_VSCROLL, SB_BOTTOM, 0);
	}
}

void CGUILog::OnAdd(const string &text)
{
	CLog::OnAdd(text);
	if(editControl)
	{
		int i = editControl->LineCount() - 1;
		editControl->SetLine(i, editControl->Line(i) + text);
	}
}

void CGUILog::OnUpdateState(const vector<LogState> &states)
{
	CLog::OnUpdateState(states);
}
