#include "System.h"
#include "Menus.h"

TCHAR MnemonicFromMessage(MSG &msg)
{
	return msg.message == WM_SYSKEYDOWN ? (TCHAR)MapVirtualKey(msg.wParam, 2)
		: msg.message == WM_SYSCHAR ? (TCHAR)msg.wParam : TEXT('\0');
}

string VirtualKeyToString(BYTE key, bool extended)
{
	const int maxVirtualKeyName = 100;
	string s(maxVirtualKeyName);
	GetKeyNameText((MapVirtualKey(key, 0) << 16) | (extended ? 0x1000000 : 0), s, maxVirtualKeyName);
	s.Update();
	return s;
}

//////////////
//          //
// Shortcut //
//          //
//////////////

const WORD Shortcut::ControlMask = 0x01 << 8;
const WORD Shortcut::ShiftMask = 0x02 << 8;
const WORD Shortcut::AltMask = 0x04 << 8;
const WORD Shortcut::KeyMask = 0xFF;
const Shortcut Shortcut::None = Shortcut();

Shortcut::Shortcut(BYTE key, bool control, bool shift, bool alt)
{
	Create(key, control, shift, alt);
}

Shortcut::Shortcut(MSG &msg)
{
	if(msg.message == WM_KEYDOWN || msg.message == WM_SYSKEYDOWN)
		Create((BYTE)msg.wParam, IsKeyPressed(VK_CONTROL), IsKeyPressed(VK_SHIFT), IsKeyPressed(VK_MENU));
	else data = 0;
}

Shortcut::Shortcut(const Shortcut &value)
{
	this->data = value.data;
}

void Shortcut::Create(BYTE key, bool control, bool shift, bool alt)
{
	if(key >= VK_BACK && key != VK_SHIFT && key != VK_CONTROL && key != VK_MENU
		&& key != VK_LSHIFT && key != VK_RSHIFT
		&& key != VK_LCONTROL && key != VK_RCONTROL
		&& key != VK_LMENU && key != VK_RMENU)
	{
		data = key;
		if(control) data |= ControlMask;
		if(shift) data |= ShiftMask;
		if(alt) data |= AltMask;
	}
	else data = 0;
}

string Shortcut::ToString()
{
	string s;
	if(Control) s = VirtualKeyToString(VK_CONTROL, false);
	if(Shift)
	{
		if(!s.IsEmpty()) s += TEXT('+');
		s += VirtualKeyToString(VK_SHIFT, false);
	}
	if(Alt)
	{
		if(!s.IsEmpty()) s += TEXT('+');
		s += VirtualKeyToString(VK_MENU, false);
	}
	if(!s.IsEmpty()) s += TEXT('+');
	BYTE key = Key;
	bool isExtKey = false;
	switch(key)
	{
		case VK_INSERT:
		case VK_DELETE:
		case VK_HOME:
		case VK_END:
		case VK_NEXT:
		case VK_PRIOR:
		case VK_NUMLOCK:
		case VK_CANCEL:
		case VK_LEFT:
		case VK_UP:
		case VK_RIGHT:
		case VK_DOWN:
		case VK_DIVIDE:
		case VK_LWIN:
		case VK_RWIN:
		case VK_APPS:
			isExtKey = true;
			break;
		case VK_PAUSE:
			key = VK_NUMLOCK;
			break;
	}
	return s + VirtualKeyToString(key, isExtKey);
}

///////////////
//           //
// CMenuItem //
//           //
///////////////

const DWORD CMenuItem::InfoAllMask = MIIM_FTYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU | MIIM_CHECKMARKS | MIIM_STRING | MIIM_BITMAP;

CMenuItem::CMenuItem()
{
	Create();
}

CMenuItem::CMenuItem(MenuItemType type, WORD id, void * value)
{
	Create(type, false, id, NULL, emptyString, value);
}

CMenuItem::CMenuItem(const string &text, WORD id, ::Shortcut shortcut)
{
	Create(mitText, false, id, NULL, text, NULL, shortcut);
}

CMenuItem::CMenuItem(const string &text, WORD id, void * value, ::Shortcut shortcut)
{
	Create(mitText, false, id, NULL, text, value, shortcut);
}

CMenuItem::CMenuItem(const string &text, bool radioCheck, WORD id, ::Shortcut shortcut)
{
	Create(mitText, radioCheck, id, NULL, text, NULL, shortcut);
}

CMenuItem::CMenuItem(const string &text, bool radioCheck, WORD id, void * value, ::Shortcut shortcut)
{
	Create(mitText, radioCheck, id, NULL, text, value, shortcut);
}

CMenuItem::CMenuItem(const string &text, CMenu * subMenu)
{
	Create(mitText, false, id, subMenu, text);
}

CMenuItem::CMenuItem(HMENU hMenu, int index)
{
	Create();
	Assign(hMenu, index);
}

CMenuItem::~CMenuItem()
{
	delete subMenu;
}

void CMenuItem::Create(MenuItemType type, bool radioCheck, WORD id, CMenu * subMenu,
	const string &text, void * value, ::Shortcut shortcut)
{
	menu = NULL;
	this->type = type;
	_break = mbNone;
	this->radioCheck = radioCheck;
	rightJustify = false;
	rightOrder = false;
	checked = false;
	_default = false;
	enabled = true;
	this->id = id;
	this->subMenu = subMenu;
	this->text = text;
	bitmap = NULL;
	this->value = value;
	checkedBitmap = NULL;
	uncheckedBitmap = NULL;
	this->shortcut = shortcut;
}

int CMenuItem::get_Index()
{
	return menu ? menu->Items->IndexOf(this) : -1;
}

void CMenuItem::set_Type(MenuItemType value)
{
	if(type != value)
	{
		type = value;
		if(type != mitText) text.Clear();
		DWORD mask = MIIM_FTYPE || MIIM_STRING;
		if(type == mitSeparator)
		{
			shortcut == Shortcut::None;
			id = 0;
			mask |= MIIM_ID;
		}
		OnChanged(mask);
	}
}

void CMenuItem::set_Break(MenuBreak value)
{
	if(_break != value)
	{
		_break = value;
		OnChanged(MIIM_FTYPE);
	}
}

void CMenuItem::set_RadioCheck(bool value)
{
	if(radioCheck != value)
	{
		radioCheck = value;
		OnChanged(MIIM_FTYPE);
	}
}

void CMenuItem::set_RightJustify(bool value)
{
	if(rightJustify != value)
	{
		rightJustify = value;
		OnChanged(MIIM_FTYPE);
	}
}

void CMenuItem::set_RightOrder(bool value)
{
	if(rightOrder != value)
	{
		rightOrder = value;
		OnChanged(MIIM_FTYPE);
	}
}

void CMenuItem::set_Checked(bool value)
{
	if(checked != value)
	{
		checked = value;
		OnChanged(MIIM_STATE);
	}
}

void CMenuItem::set_Default(bool value)
{
	if(_default != value)
	{
		_default = value;
		OnChanged(MIIM_STATE);
	}
}

void CMenuItem::set_Enabled(bool value)
{
	if(enabled != value)
	{
		enabled = value;
		OnChanged(MIIM_STATE);
	}
}

void CMenuItem::set_Id(WORD value)
{
	if(type == mitSeparator) value = 0;
	if(id != value)
	{
		id = value;
		OnChanged(MIIM_ID);
	}
}

void CMenuItem::set_SubMenu(CMenu * value)
{
	if(subMenu != value)
	{
		subMenu = value;
		DWORD mask = MIIM_SUBMENU;
		if(subMenu && !shortcut.IsNone)
		{
			shortcut == Shortcut::None;
			mask |= MIIM_STRING;
		}
		OnChanged(mask);
	}
}

void CMenuItem::set_Text(const string & value)
{
	string v = type == mitText ? value : emptyString;
	if(text != v)
	{
		text = v;
		OnChanged(MIIM_STRING);
	}
}

void CMenuItem::set_Bitmap(HBITMAP value)
{
	if(bitmap != value)
	{
		bitmap = value;
		OnChanged(MIIM_BITMAP);
	}
}

void CMenuItem::set_Value(void * value)
{
	if(this->value != value)
	{
		this->value = value;
		if(type == mitOwnerDraw) OnChanged(MIIM_FTYPE);
	}
}

void CMenuItem::set_CheckedBitmap(HBITMAP value)
{
	if(checkedBitmap != value)
	{
		checkedBitmap = value;
		OnChanged(MIIM_CHECKMARKS);
	}
}

void CMenuItem::set_UncheckedBitmap(HBITMAP value)
{
	if(uncheckedBitmap != value)
	{
		uncheckedBitmap = value;
		OnChanged(MIIM_CHECKMARKS);
	}
}

void CMenuItem::set_Shortcut(::Shortcut value)
{
	if(type == mitSeparator) value = Shortcut::None;
	if(shortcut != value)
	{
		shortcut = value;
		OnChanged(MIIM_STRING);
	}
}

void CMenuItem::OnChanged(DWORD mask)
{
	Update(mask);
	if(menu) menu->OnChanged();
}

void CMenuItem::Update(DWORD mask)
{
	if(!menu) return;
	MENUITEMINFO mii;
	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = 0;
	if(mask & MIIM_FTYPE)
	{
		mii.fMask |= MIIM_FTYPE;
		mii.fType = type | _break;
		if(radioCheck) mii.fType |= MFT_RADIOCHECK;
		if(rightJustify) mii.fType |= MFT_RIGHTJUSTIFY;
		if(rightOrder) mii.fType |= MFT_RIGHTORDER;
	}
	if(mask & MIIM_STATE)
	{
		mii.fMask |= MIIM_STATE;
		mii.fState =
				(checked ? MFS_CHECKED : MFS_UNCHECKED)
			| (_default ? MFS_DEFAULT : 0)
			| (enabled ? MFS_ENABLED : MFS_DISABLED);
	}
	if(mask & MIIM_ID)
	{
		mii.fMask |= MIIM_ID;
		mii.wID = id;
	}
	if(mask & MIIM_SUBMENU)
	{
		mii.fMask |= MIIM_SUBMENU;
		mii.hSubMenu = subMenu ? subMenu->Handle : NULL;
	}
	string s;
	if(type == mitOwnerDraw && (mask & (MIIM_FTYPE | MIIM_STRING))) mii.dwTypeData = (LPSTR)this->value;
	else if(type != mitOwnerDraw && (mask & MIIM_STRING))
	{
		mii.fMask |= MIIM_STRING;
		s = shortcut.IsNone || text.IndexOf(TEXT('\t')) != -1
			? text
			: text + TEXT('\t') + shortcut.ToString();
		mii.dwTypeData = s;
	}
	if(mask & MIIM_CHECKMARKS)
	{
		mii.fMask |= MIIM_CHECKMARKS;
		mii.hbmpChecked = checkedBitmap;
		mii.hbmpUnchecked = uncheckedBitmap;
	}
#if(WINVER >= 0x0500)
	if(mask & MIIM_BITMAP)
	{
		mii.fMask |= MIIM_BITMAP;
		mii.hbmpItem = bitmap;
	}
#endif
	SetMenuItemInfo(menu->Handle, Index, TRUE, &mii);
}

void CMenuItem::Assign(HMENU hMenu, int index)
{
	MENUITEMINFO mii;
	mii.cbSize = sizeof(mii);
	mii.fMask = InfoAllMask;
	mii.dwTypeData = NULL;
	mii.cch = 0;
	if(!GetMenuItemInfo(hMenu, index, TRUE, &mii)) throw exception("Can not retrieve MENUITEMINFO");
	type = mii.fType & MFT_OWNERDRAW ? mitOwnerDraw : mii.fType & MFT_SEPARATOR ? mitSeparator : mitText;
	_break = (mii.fType & MFT_MENUBARBREAK) != 0 ? mbBarBreak : (mii.fType & MFT_MENUBREAK) != 0 ? mbBreak : mbNone;
	radioCheck = (mii.fType & MFT_RADIOCHECK) != 0;
	rightJustify = (mii.fType & MFT_RIGHTJUSTIFY) != 0;
	rightOrder = (mii.fType & MFT_RIGHTORDER) != 0;
	checked = (mii.fState & MFS_CHECKED) != 0;
	_default = (mii.fState & MFS_DEFAULT) != 0;
	enabled = (mii.fState & MFS_DISABLED) == 0;
	id = (WORD)mii.wID;
	HMENU hSubMenu = mii.hSubMenu;
#if(WINVER >= 0x0500)
	bitmap = mii.hbmpItem;
#else
	bitmap = NULL;
#endif
	checkedBitmap = mii.hbmpChecked;
	uncheckedBitmap = mii.hbmpUnchecked;
	if(type == mitOwnerDraw) value = mii.dwTypeData;
	else if(type == mitText)
	{
		mii.fMask = MIIM_STRING;
		text.SetLength(mii.cch);
		mii.dwTypeData = text;
		mii.cch++;
		if(!GetMenuItemInfo(hMenu, index, TRUE, &mii)) throw exception("Can not retrieve MENUITEMINFO for string");
	}
	else text.Clear();
	if(hSubMenu)
	{
		shortcut = Shortcut::None;
		delete subMenu;
		subMenu = new CMenu(hSubMenu);
	}
	OnChanged(InfoAllMask);
}

bool CMenuItem::IsMnemonic(TCHAR ch)
{
	ch = ToUpper(ch);
	for(const TCHAR * i = text.Begin(); i < text.End(); i++)
		if(*i == TEXT('&'))
		{
			i++;
			if(i == text.End()) break;
			else if(*i != TEXT('&') && ToUpper(*i) == ch) return true;
		}
	return false;
}

/////////////////////////
//                     //
// CMenuItemCollection //
//                     //
/////////////////////////

CMenuItemCollection::CMenuItemCollection(CMenu * owner)
	: CCollection<CMenuItem>(true), owner(owner)
{
	if(!owner) throw exception("MenuItemCollection owner is NULL");
}

void CMenuItemCollection::OnInsertComplete(int index, CMenuItem * value)
{
	CCollection<CMenuItem>::OnInsertComplete(index, value);
	owner->OnMenuItemAdded(index, value);
}

void CMenuItemCollection::OnRemoveComplete(int index, CMenuItem * value)
{
	owner->OnMenuItemRemoved(index, value);
	CCollection<CMenuItem>::OnRemoveComplete(index, value);
}

void CMenuItemCollection::OnAdded(CMenuItem * value)
{
	CCollection<CMenuItem>::OnAdded(value);
	value->menu = owner;
}

void CMenuItemCollection::OnRemoved(CMenuItem * value)
{
	value->menu = NULL;
	CCollection<CMenuItem>::OnRemoved(value);
}

void CMenuItemCollection::OnValidate(CMenuItem * value)
{
	CCollection<CMenuItem>::OnValidate(value);
	if(value->menu) throw exception("MenuItem already belongs to a Menu");
}

///////////
//       //
// CMenu //
//       //
///////////

CMenu::CMenu()
{
	Create();
}

CMenu::CMenu(HMENU hMenu)
{
	Create();
	try
	{
		Assign(hMenu);
	}
	catch(...)
	{
		Destroy();
		throw;
	}
}

CMenu::CMenu(LPCTSTR resourceName)
{
	Create();
	try
	{
		Assign(resourceName);
	}
	catch(...)
	{
		Destroy();
		throw;
	}
}

CMenu::~CMenu()
{
	Destroy();
}

void CMenu::Create()
{
	handle = CreateMenu();
	if(!handle) throw exception("Can not create HMENU");
	items = new CMenuItemCollection(this);
}

void CMenu::Destroy()
{
	delete items;
	if(!DestroyMenu(handle))
		throw exception("Can not destroy HMENU");
}

void CMenu::OnMenuItemAdded(int index, CMenuItem * value)
{
	if(!InsertMenu(handle, index, MF_BYPOSITION, 0, emptyString))
		throw exception("Can not insert to HMENU");
	value->Update();
}

void CMenu::OnMenuItemRemoved(int index, CMenuItem * /*value*/)
{
	if(!RemoveMenu(handle, index, MF_BYPOSITION)) throw exception("Can not remove from HMENU");
}

void CMenu::OnChanged()
{
}

CMenuItem * CMenu::GetMenuItemWithMnemonic(TCHAR ch)
{
	if(ch == TEXT('\0')) return NULL;
	for(int i = 0; i < items->Count; i++)
		if(items->Item[i]->IsMnemonic(ch)) return items->Item[i];
	return NULL;
}

bool CMenu::IsMnemonic(TCHAR ch)
{
	return GetMenuItemWithMnemonic(ch) != NULL;
}

CMenuItem * CMenu::GetMenuItemWithShortcut(Shortcut value)
{
	if(value.IsNone) return NULL;
	int i;
	for(i = 0; i < items->Count; i++)
		if(items->Item[i]->Shortcut == value) return items->Item[i];
	for(i = 0; i < items->Count; i++)
		if(items->Item[i]->HasSubMenu)
		{
			CMenuItem * item = items->Item[i]->SubMenu->GetMenuItemWithShortcut(value);
			if(item) return item;
		}
	return NULL;
}

CMenuItem * CMenu::GetMenuItemWithId(WORD value)
{
	if(value == 0) return NULL;
	int i;
	for(i = 0; i < items->Count; i++)
		if(items->Item[i]->Id == value) return items->Item[i];
	for(i = 0; i < items->Count; i++)
		if(items->Item[i]->HasSubMenu)
		{
			CMenuItem * item = items->Item[i]->SubMenu->GetMenuItemWithId(value);
			if(item) return item;
		}
	return NULL;
}

void CMenu::Assign(HMENU value)
{
	int count = GetMenuItemCount(value);
	for(int i = 0; i < count; i++)
		items->Add(new CMenuItem(value, i));
}

void CMenu::Assign(LPCTSTR resourceName)
{
	HMENU hMenu = LoadMenu(Application()->Handle(), resourceName);
	if(!hMenu) throw exception("Can not load HMENU from resources");
	Assign(hMenu);
	if(!DestroyMenu(hMenu)) throw exception("Can not destroy HMENU loaded from resources");
}

void CMenu::AssignAccelerators(HACCEL value)
{
	vector<ACCEL> accels(CopyAcceleratorTable(value, NULL, 0));
	if(CopyAcceleratorTable(value, accels.Data(), accels.Length()) != accels.Length())
		throw exception("Can not retrieve accelerator table");
	for(ACCEL * pAccel = accels.Begin(); pAccel < accels.End(); pAccel++)
	{
		CMenuItem * item = GetMenuItemWithId(pAccel->cmd);
		item->Shortcut = pAccel->fVirt & FVIRTKEY
			? Shortcut((BYTE)pAccel->key, (pAccel->fVirt & FCONTROL) != 0, (pAccel->fVirt & FSHIFT) != 0, (pAccel->fVirt & FALT) != 0)
			: Shortcut::None;
	}
}

void CMenu::AssignAccelerators(LPCTSTR resourceName)
{
	HACCEL hAccel = LoadAccelerators(Application()->Handle(), resourceName);
	if(!hAccel) throw exception("Can not load HACCEL from resources");
	AssignAccelerators(hAccel);
	if(!DestroyAcceleratorTable(hAccel)) throw exception("Can not destroy HACCEL loaded from resources");
}
