#pragma once

#ifndef REG_QWORD
#define REG_QWORD 11
#endif

class CRegistry
{
private:
	HKEY root_key;
	HKEY key;
	string key_name;

	void CheckValueIndex(int index) { if(index < 0 || index >= ValueCount()) throw exception("Invalid registry value index"); }

public:
	CRegistry(HKEY root_key);
	~CRegistry();

	HKEY RootKey() { return root_key; }

	HKEY Key() { return key ? key : root_key; }
	void OpenKey(const string &name, bool write = false, bool create = false);
	void CloseKey();
	void FlushKey();
	const string & KeyName() { return key_name; }

	void DeleteSubKey(const string &name);
	void GetSubKeyNames(strings &names);
	int SubKeyCount();
	string SubKeyName(int index);
	int SubKeyIndex(const string &name);
	bool SubKeyExists(const string &name);

	void DeleteValue(const string &name);
	void GetValueNames(strings &names);
	int ValueCount();
	string ValueName(int index);
	DWORD ValueType(int index);
	DWORD ValueType(const string &name);
	DWORD ValueSize(int index);
	DWORD ValueSize(const string &name);
	void GetValue(int index, BYTE *data, DWORD &size, DWORD *type = NULL);
	void GetValue(const string &name, BYTE *data, DWORD &size, DWORD *type = NULL);
	int ValueIndex(const string &name);
	bool ValueExists(const string &name);
	DWORD GetDWordValue(const string &name);
	QWORD GetQWordValue(const string &name);
	int GetIntValue(const string &name);
	bool GetBoolValue(const string &name);
	float GetFloatValue(const string &name);
	double GetDoubleValue(const string &name);
	string GetStringValue(const string &name);
	DWORD GetDWordValue(const string &name, DWORD default_value);
	QWORD GetQWordValue(const string &name, QWORD default_value);
	int GetIntValue(const string &name, int default_value);
	bool GetBoolValue(const string &name, bool default_value);
	float GetFloatValue(const string &name, float default_value);
	double GetDoubleValue(const string &name, double default_value);
	string GetStringValue(const string &name, const string &default_value);
	void SetValue(const string &name, const BYTE *data, DWORD size, DWORD type = REG_BINARY);
	void SetDWordValue(const string &name, DWORD value);
	void SetQWordValue(const string &name, QWORD value);
	void SetIntValue(const string &name, int value);
	void SetBoolValue(const string &name, bool value);
	void SetFloatValue(const string &name, float value);
	void SetDoubleValue(const string &name, double value);
	void SetStringValue(const string &name, const string &value);

};
