#pragma once

#include "Controls.h"
#include "Utils.h"

int GetDefaultControlHeight(LPCTSTR controlClass, DWORD style = 0, DWORD exStyle = 0);

enum TextAlignment { taDefault = -1, taLeft, taCenter, taRight };
enum TextEllipsis { teNone, teEnd, teWord, tePath };
enum BorderStyle { bsDefault = -1, bsNone, bsRaised, bsLowered };

class CPanel: public CControl
{
protected:
	LRESULT OnGetDlgCode(WPARAM /*wParam*/, MSG * /*message*/) { return DLGC_STATIC; }

public:
	CPanel();
};

class CCaptionPanel: public CPanel
{
private:
	bool active;
	CControl *active_control;

	void UpdateActive();
	void SelectTheControl();
	bool SelectAnyControl(CControl * control);

protected:
	LRESULT OnGetDlgCode(WPARAM /*wParam*/, MSG * /*message*/) { return DLGC_STATIC; }
	LRESULT OnMessage(UINT message, WPARAM wParam, LPARAM lParam);
	void OnActiveChanged();
	void OnActiveControlChanged();
	void OnControlParentChanged(CControl * control);

	bool Active() { return active; }
	CControl * ActiveControl() { return active_control; }
	void SetActive(bool active);
	void SetActiveControl(CControl * control);

public:
	CCaptionPanel();
};

#define SN_SPLITTED 2

class CSplitter: public CControl
{
private:
	bool splitting;
	int x0, y0;

	float ratio;
	CControl *prev_control, *next_control;

protected:
	LRESULT OnGetDlgCode(WPARAM /*wParam*/, MSG * /*message*/) { return DLGC_STATIC; }

	LRESULT OnCancelMode(WPARAM wParam, LPARAM lParam);
	LRESULT OnLButtonDown(DWORD keys, int x, int y);
	LRESULT OnLButtonUp(DWORD keys, int x, int y);
	LRESULT OnMouseMove(DWORD keys, int x, int y);
	virtual void OnSplit(int x, int y) = 0;

public:
	CSplitter();

	float Ratio() { return ratio; }
	CControl * PrevControl() { return prev_control; }
	CControl * NextControl() { return next_control; }

	void SetRatio(float value);
	void SetPrevControl(CControl * value);
	void SetNextControl(CControl * value);

};

class CHSplitter: public CSplitter
{
protected:
	HCURSOR GetCursor() { return LoadCursor(GetModuleHandle(TEXT("comctl32.dll")), MAKEINTRESOURCE(135)); }
	
	void OnSplit(int x, int y);

public:
	CHSplitter();

	int AreaTop();
	int AreaBottom();
	int AreaHeight();
	int TopHeight();
	int BottomHeight();
	void UpdateControls(int x, int width, CControl * top_control, CControl * bottom_control);

};

class CVSplitter: public CSplitter
{
protected:
	HCURSOR GetCursor() { return LoadCursor(GetModuleHandle(TEXT("comctl32.dll")), MAKEINTRESOURCE(107)); }

	void OnSplit(int x, int y);

public:
	CVSplitter();

	int AreaLeft();
	int AreaRight();
	int AreaWidth();
	int LeftWidth();
	int RightWidth();
	void UpdateControls(int y, int height, CControl * left_control, CControl * right_control);

};

class CStatic: public CControl
{
protected:
	string GetWndClassName() { return TEXT("STATIC"); }

	DWORD GetWndStyle();
	virtual DWORD GetStaticStyle() = 0;

public:
	CStatic();
};

#ifndef SS_EDITCONTROL
#define SS_EDITCONTROL 0x00002000L
#endif

enum LabelStyle { lsStandard = 0, lsSimple = SS_SIMPLE, lsEditControl = SS_EDITCONTROL };

class CLabel: public CStatic
{
private:
	LabelStyle style;
	TextAlignment textAlignment;
	TextEllipsis ellipsis;
	bool wordWrap;
	bool noPrefix;

protected:
	DWORD GetStaticStyle();

public:
	CLabel();

	LabelStyle Style() { return style; }
	TextAlignment TextAlignment() { return textAlignment; }
	TextEllipsis Ellipsis() { return ellipsis; }
	bool WordWrap() { return wordWrap; }
	bool NoPrefix() { return noPrefix; }

	void SetStyle(LabelStyle value);
	void SetTextAlignment(::TextAlignment value);
	void SetEllipsis(TextEllipsis value);
	void SetWordWrap(bool value);
	void SetNoPrefix(bool value);

};

class CPicture: public CStatic
{
private:
	HBITMAP bitmap;

	void UpdateBitmap();

protected:
	DWORD GetStaticStyle();

	void OnCreate();

public:
	CPicture();

	HBITMAP Bitmap() { return bitmap; }
	void SetBitmap(HBITMAP value);
};

enum SimpleGraphicsStyle
{
	sgsBlackRect = SS_BLACKRECT, sgsGrayRect = SS_GRAYRECT, sgsWhiteRect = SS_BLACKRECT,
	sgsBlackFrame = SS_BLACKFRAME, sgsGrayFrame = SS_GRAYFRAME, sgsWhiteFrams = SS_WHITEFRAME,
	sgsEtchedHorz = SS_ETCHEDHORZ, sgsEtchedVert = SS_ETCHEDVERT, sgsEtchedFrame = SS_ETCHEDFRAME
};

class CSimpleGraphics: public CStatic
{
private:
	SimpleGraphicsStyle style;

protected:
	DWORD GetStaticStyle();

public:
	CSimpleGraphics();

	SimpleGraphicsStyle Style() { return style; }
	void SetStyle(SimpleGraphicsStyle value);
};

class CButtonControl: public CControl
{
protected:
	string GetWndClassName() { return TEXT("BUTTON"); }

	DWORD GetWndStyle();
	virtual DWORD GetButtonStyle() = 0;

	void Click();
	void OnCommandSent(WORD event);
	virtual void OnClick() {}

public:
	CButtonControl();
};

class CButton: public CButtonControl
{
private:
	bool def;
	bool cancel;
	int modal_result;
	bool changing_button_style;

	void UpdateButtonStyle();

protected:
	DWORD GetButtonStyle();

	void OnClick();
	bool OnDefault();
	bool OnCancel();
	void OnActiveControlChanged();
	void OnActiveChanged();
	LRESULT OnMessage(UINT message, WPARAM wParam, LPARAM lParam);

public:
	CButton();

	bool Default() { return def; }
	bool Cancel() { return cancel; }
	void SetDefault(bool value);
	void SetCancel(bool cancel);
	int ModalResult() { return modal_result; }
	void SetModalResult(int value);
};

class CCheckBox: public CButtonControl
{
private:
	bool checked;

	void UpdateChecked();

protected:
	DWORD GetButtonStyle();

	void OnCreate();
	void OnClick();

public:
	CCheckBox();

	bool Checked() { return checked; }
	void SetChecked(bool value);

	void Toggle();

};

class CRadioButton: public CButtonControl
{
private:
	bool checked;

	void UpdateChecked();

protected:
	DWORD GetButtonStyle();

	void OnCreate();
	void OnClick();

public:
	CRadioButton();

	void SetParent(CControl *value);

	bool Checked() { return checked; }
	void SetChecked(bool value);

	void Check();

};

class CGroupBox: public CButtonControl
{
protected:
	void OnCreate();
	DWORD GetButtonStyle();

	LRESULT OnEraseBackground(HDC dc, LPARAM lParam);

public:
	CGroupBox();
};

class CEdit: public CControl
{
private:
	bool changing_text;
	bool multi_line;
	bool read_only;
	bool horz_scroll_bar;
	bool vert_scroll_bar;
	bool word_wrap;
	TCHAR password_char;
	int limit_text;

	strings lines;

	void UpdateLines();

	void CheckIndex(int index, bool end = false)
	{
		if(index < 0 || (end && index > lines.Length()) || (!end && index >= lines.Length()))
			throw exception("Edit line index out of bounds");
	}

	void DoTextChanged();

protected:
	string GetWndClassName() { return TEXT("EDIT"); }

	DWORD GetWndStyle();

	void OnCreate();
	void OnCommandSent(WORD event);
	LRESULT OnSetFocus(HWND hwnd, LPARAM lParam);
	LRESULT OnGetDlgCode(WPARAM wParam, MSG *message);

public:
	CEdit();

	bool MultiLine() { return multi_line; }
	bool ReadOnly() { return read_only; }
	bool HorzScrollBar() { return horz_scroll_bar; }
	bool VertScrollBar() { return vert_scroll_bar; }
	bool WordWrap() { return word_wrap; }

	void SetMultiLine(bool value);
	void SetReadOnly(bool value);
	void SetHorzScrollBar(bool value);
	void SetVertScrollBar(bool value);
	void SetWordWrap(bool value);

	int LineCount();
	int AddLine(const string & line)
	{
		int i = lines.Length();
		if(i == 1 && text.IsEmpty()) i = 0;
		InsertLine(i, line);
		return i;
	}
	void InsertLine(int index, const string & line);
	void RemoveLine(int index);
	void ClearLines();
	const string & Line(int index);
	void SetLine(int index, const string & value);
	void SetText(const string & value);

	bool CanUndo();
	void ClearUndo();
	void Undo();
	bool Modified();
	void SetModified(bool value);
	int SelStart();
	int SelEnd();
	int SelLength();
	void SetSelStart(int value);
	void SetSelEnd(int value);
	void SetSelLength(int value);
	void SelectAll();
	void SelectNone();
	string SelText();
	void SetSelText(const string & value);
	void ClearSelection();
	void CopyToClipboard();
	void CutToClipboard();
	void PasteFromClipboard();
	TCHAR PasswordChar();
	void SetPasswordChar(TCHAR value);
	int LimitText();
	void SetLimitText(int value);
};

enum ListBoxSelStyle { lbsNone = LBS_NOSEL, lbsSingle = 0, lbsMultiple = LBS_MULTIPLESEL, lbsExtended = LBS_EXTENDEDSEL };

class CListBox: public CControl
{
private:
	ListBoxSelStyle sel_style;

	strings items;

	void UpdateItems();

	void CheckIndex(int index, bool end = false)
	{
		if(index < 0 || (end && index > items.Length()) || (!end && index >= items.Length()))
			throw exception("List box item index out of bounds");
	}

protected:
	string GetWndClassName() { return TEXT("LISTBOX"); }

	DWORD GetWndStyle();

	void OnCreate();

public:
	CListBox();

	ListBoxSelStyle SelStyle() { return sel_style; }
	void SetSelStyle(ListBoxSelStyle value);

	int ItemCount();
	int AddItem(const string & value)
	{
		int i = items.Length();
		InsertItem(i, value);
		return i;
	}
	void RemoveItem(int index);
	void InsertItem(int index, const string & value);
	void ClearItems();
	const string & Item(int index);
	void SetItem(int index, const string & value);
	int SelIndex();
	void SetSelIndex(int index);
	int SelCount();
	bool ItemSelected(int index);
	void SetItemSelected(int index, bool value);

	void SelectAll();
	void SelectNone();
};

enum ComboBoxStyle { cbsSimple = CBS_SIMPLE, cbsDropDown = CBS_DROPDOWN, cbsDropDownList = CBS_DROPDOWNLIST };

class CComboBox: public CControl
{
private:
	ComboBoxStyle style;
	strings items;
	vector<void*> values;
	int item_index;

	void UpdateItems();
	void UpdateList();
	void UpdateItemIndex();

	void CheckIndex(int index, bool end = false)
	{
		if(index < 0 || (end && index > items.Length()) || (!end && index >= items.Length()))
			throw exception("Combo box item index out of bounds");
	}

protected:
	string GetWndClassName() { return TEXT("COMBOBOX"); }

	DWORD GetWndStyle();

	void OnCreate();

	void OnCommandSent(WORD event);

public:
	CComboBox();

	ComboBoxStyle Style() { return style; }
	void SetStyle(ComboBoxStyle value);

	int ItemCount();
	void ClearItems();
	int AddItem(const string &text, void * value = NULL)
	{
		int i = items.Length();
		InsertItem(i, text, value);
		return i;
	}
	void InsertItem(int index, const string &text, void * value = NULL);
	void RemoveItem(int index);
	const string & Item(int index);
	void SetItem(int index, const string &value);
	void * Value(int index);
	void SetValue(int index, void * value);

	void SetText(const string &value);

	int ItemIndex();
	void SetItemIndex(int index);

	int IndexOfItem(const string &value);
	int IndexOfValue(void * value);

};

class CRichEdit: public CEdit
{
protected:
	string GetWndClassName() { return RICHEDIT_CLASS; }

	void OnCreate();

public:
	CRichEdit();
};

// ListView control, report view only
struct ListViewColumn
{
	string Text;
	int Width;
	TextAlignment Align;

	ListViewColumn()
	{
		Assign(emptyString, -1, taDefault);
	}

	ListViewColumn(const string & text, int width = -1, TextAlignment align = taDefault)
	{
		Assign(text, width, align);
	}

	ListViewColumn(const ListViewColumn & column)
	{
		Assign(column.Text, column.Width, column.Align);
	}

	ListViewColumn & operator=(const ListViewColumn & column)
	{
		Assign(column);
		return *this;
	}

	void Assign(const ListViewColumn & column)
	{
		Assign(column.Text, column.Width, column.Align);
	}

	void Assign(const string & text, int width, TextAlignment align)
	{
		Text = text;
		Width = width;
		Align = align;
	}
};

typedef vector<ListViewColumn> ListViewColumns;

struct ListViewItem
{
	string Text;
	strings Subitems;

	ListViewItem()
	{
	}

	ListViewItem(const string & text)
	{
		Assign(text);
	}

	ListViewItem(const string & text, const strings & subitems)
	{
		Assign(text, subitems);
	}

	ListViewItem(const ListViewItem & item)
	{
		Assign(item);
	}

	ListViewItem & operator=(const ListViewItem & item)
	{
		Assign(item);
		return *this;
	}

	void Assign(const string & text)
	{
		Assign(text, strings());
	}

	void Assign(const string & text, const strings & subitems)
	{
		Text = text;
		Subitems = subitems;
	}

	void Assign(const ListViewItem & item)
	{
		Assign(item.Text, item.Subitems);
	}

};

typedef vector<ListViewItem> ListViewItems;

class CListView: public CControl
{
private:
	bool owner_data;
	ListViewColumns columns;
	bool no_column_header;
	bool no_sort_header;
	bool show_sel_always;
	bool single_sel;
	bool full_row_select;
	ListViewItems items;
	int itemCount;

	void UpdateColumns();
	void UpdateItems();

	void CheckColumnIndex(int index, bool end = false)
	{
		if(index < 0 || (end && index > columns.Length()) || (!end && index >= columns.Length()))
			throw exception("List view column index out of bounds");
	}

	void CheckIndex(int index, bool end = false)
	{
		int c = owner_data ? itemCount : items.Length();
		if(index < 0 || (end && index > c) || (!end && index >= c))
			throw exception("List view item index out of bounds");
	}

	void CheckSubIndex(int index, int subindex, bool end = false)
	{
		CheckIndex(index);
		if(index < 0 || (end && subindex > columns.Length()) || (!end && subindex >= columns.Length()))
			throw exception("List view subitem index out of bounds");
	}

protected:
	string GetWndClassName() { return WC_LISTVIEW; }

	DWORD GetWndStyle();

public:
	CListView();

	void OnCreate();

	bool OwnerData() { return owner_data; }
	void SetOwnerData(bool value);

	int ColumnCount();
	void ClearColumns();
	int AddColumn(const string & text, int width = -1, TextAlignment align = taDefault)
	{
		int i = columns.Length();
		InsertColumn(i, text, width, align);
		return i;
	}
	void InsertColumn(int index, const string & text, int width = -1, TextAlignment align = taDefault);
	void RemoveColumn(int index);

	bool NoColumnHeader() { return no_column_header; }
	bool NoSortHeader() { return no_sort_header; }
	bool ShowSelAlways() { return show_sel_always; }
	bool SingleSel() { return single_sel; }
	bool FullRowSelect() { return full_row_select; }
	void SetNoColumnHeader(bool value);
	void SetNoSortHeader(bool value);
	void SetShowSelAlways(bool value);
	void SetSingleSel(bool value);
	void SetFullRowSelect(bool value);

	int ItemCount();
	void SetItemCount(int value);
	void ClearItems();
	int AddItem(const string & text = emptyString)
	{
		int i = items.Length();
		InsertItem(i, text);
		return i;
	}
	void InsertItem(int index, const string & text = emptyString);
	void RemoveItem(int index);
	const string & ItemText(int index);
	const string & SubItemText(int index, int subindex);
	void SetItemText(int index, const string & text);
	void SetSubItemText(int index, int subindex, const string & text);

	int FocusIndex();
	void SetFocusIndex(int value);
	int SelIndex();
	void SetSelIndex(int value);
	int SelCount();
	bool ItemSelected(int index);
	void SetItemSelected(int index, bool value);
	void SelectAll();
	void SelectNone();

	void EnsureVisible(int index, bool partialOK = false);

	RECT ItemRect(int index, int code = LVIR_BOUNDS);
	RECT SubItemRect(int index, int subIndex, int code = LVIR_BOUNDS);

	void RedrawItems(int index, int count = 1);
	void RedrawItems();
};

class CProgressBar: public CControl
{
private:
	int min;
	int max;
	int position;

	void UpdateRange();
	void UpdatePosition();

protected:
	string GetWndClassName() { return PROGRESS_CLASS; }

public:
	CProgressBar();

	int Min() { return min; }
	int Max() { return max; }
	int Position() { return position; }

	void SetMin(int value);
	void SetMax(int value);
	void SetPosition(int value);
};

///////////////
//           //
// CTrackBar //
//           //
///////////////

class CTrackBar: public CControl
{
private:
	int min;
	int max;
	int position;
	int line_size;
	int page_size;

	CControl *first_buddy;
	CControl *last_buddy;

	void UpdateRange();
	void UpdatePosition();
	void UpdateLineSize();
	void UpdatePageSize();
	void UpdateFirstBuddy();
	void UpdateLastBuddy();

protected:
	string GetWndClassName() { return TRACKBAR_CLASS; }

	void OnCreate();

	void OnHScrollSent(WORD code, WORD pos);
	void OnVScrollSent(WORD code, WORD pos);

public:
	CTrackBar();

	int Min() { return min; }
	int Max() { return max; }
	int Position() { return position; }
	int LineSize() { return line_size; }
	int PageSize() { return page_size; }
	CControl * FirstBuddy() { return first_buddy; }
	CControl * LastBuddy() { return last_buddy; }

	void SetMin(int value);
	void SetMax(int value);
	void SetPosition(int value);
	void SetLineSize(int value);
	void SetPageSize(int value);
	void SetFirstBuddy(CControl * value);
	void SetLastBuddy(CControl * value);
};

////////////////
//            //
// CStatusBar //
//            //
////////////////

struct StatusBarPart
{
	int Width;
	string Text;
	TextAlignment Align;
	BorderStyle Border;

	StatusBarPart()
	{
		Assign(50, emptyString, taDefault, bsDefault);
	}

	StatusBarPart(int width, const string &text, TextAlignment align = taDefault, BorderStyle border = bsDefault)
	{
		Assign(width, text, align, border);
	}

	StatusBarPart(const StatusBarPart &part)
	{
		Assign(part.Width, part.Text, part.Align, part.Border);
	}

	StatusBarPart& operator=(const StatusBarPart &part)
	{
		Assign(part);
		return *this;
	}

	void Assign(const StatusBarPart &part)
	{
		Assign(part.Width, part.Text, part.Align, part.Border);
	}

	void Assign(int width, const string &text, TextAlignment align, BorderStyle border)
	{
		Width = width;
		Text = text;
		Align = align;
		Border = border;
	}
};

typedef vector<StatusBarPart> StatusBarParts;

class CStatusBar: public CControl
{
private:
	StatusBarParts parts;
	bool size_grip;
	bool simple;
	TextAlignment text_align;
	BorderStyle text_border;

	void UpdatePart(int index);
	void UpdateParts();
	void UpdateSimple();

	void CheckIndex(int index, bool end = false)
	{
		if(index < 0 || (end && index > parts.Length()) || (!end && index >= parts.Length()))
			throw exception("Status bar part index out of bounds");
	}

protected:
	string GetWndClassName() { return STATUSCLASSNAME; }
	DWORD GetWndStyle();

	void OnCreate();

public:
	CStatusBar();

	bool SizeGrip();
	void SetSizeGrip(bool value);

	bool Simple();
	void SetSimple(bool value);

	int PartCount();
	const string & PartText(int index);
	TextAlignment PartAlign(int index);
	BorderStyle PartBorder(int index);
	int PartWidth(int index);
	void SetPartText(int index, const string & value);
	void SetPartAlign(int index, TextAlignment value);
	void SetPartBorder(int index, BorderStyle value);
	void SetPartWidth(int index, int value);
	int AddPart(int width, const string & text = emptyString, TextAlignment align = taDefault, BorderStyle border = bsDefault)
	{
		int i = parts.Length();
		InsertPart(i, width, text, align, border);
		return i;
	}
	void InsertPart(int index, int width, const string & text = emptyString, TextAlignment align = taDefault, BorderStyle border = bsDefault);
	void RemovePart(int index);
	void ClearParts();

	void SetText(const string & value);
	TextAlignment TextAlign() { return text_align; }
	BorderStyle TextBorder() { return text_border; }
	void SetTextAlign(TextAlignment value);
	void SetTextBorder(BorderStyle value);
};

class CImageList
{
private:
	HIMAGELIST handle;
	int width;
	int height;

	void Init(int width, int height);

public:
	CImageList();
	CImageList(int width, int height);
	~CImageList();

	HIMAGELIST Handle() { return handle; }
	int Width() { return width; }
	int Height() { return height; }
	int Count();

	int Add(HICON icon);
	int Add(HBITMAP bitmap);
	int Add(HBITMAP bitmap, COLORREF maskColor);
};

class CToolBar: public CControl
{
public:
	CToolBar();
};
