unit FormWait;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Math, AppEvnts;

const
  UM_FIRST_SHOW = WM_USER;

type
  TWaitForm = class(TForm)
    ApplicationEvents: TApplicationEvents;
    procedure FormPaint(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormShow(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
  private
    ShowedOnce: Boolean;
    FMessage: string;
    FThread: TThread;
    procedure UMFirstShow(var Message: TMessage); message UM_FIRST_SHOW;
    procedure FormFirstShow(Sender: TObject);
    procedure SetMessage(const Value: string);
    procedure SetThread(const Value: TThread);
  public
    constructor Create(AOwner: TComponent); override;
    property Thread: TThread read FThread write SetThread;
    property Message: string read FMessage write SetMessage;
  end;

procedure WaitBox(Owner: TForm; Thread: TThread; Message: string);

implementation

{$R *.dfm}

procedure WaitBox(Owner: TForm; Thread: TThread; Message: string);
var
  Form: TWaitForm;
begin
  Form := TWaitForm.Create(Owner);
  Form.Thread := Thread;
  Form.Message := Message;
  Form.ShowModal;
  Form.Free;
end;

{ TWaitForm }

procedure TWaitForm.SetMessage(const Value: string);
begin
  FMessage := Value;
  Canvas.Font.Assign(Font);
  ClientWidth := Max(Canvas.TextWidth(FMessage) + 16, 150);
  ClientHeight := Max(Canvas.TextHeight(FMessage) + 16, 70);
end;

procedure TWaitForm.SetThread(const Value: TThread);
begin
  FThread := Value;
end;

procedure TWaitForm.FormPaint(Sender: TObject);
var
  OldBkMode: Integer;
  Rect: TRect;
  X, Y: Integer;
begin
  Canvas.Font.Assign(Font);
  OldBkMode := SetBkMode(Canvas.Handle, TRANSPARENT);
  Rect := ClientRect;
  X := Canvas.TextWidth(FMessage);
  Y := Canvas.TextHeight(FMessage);
  Rect.Left := (Rect.Right - X) div 2;
  Rect.Top := (Rect.Bottom - Y) div 2;
  Rect.Right := Rect.Left + X;
  Rect.Bottom := Rect.Top + Y;
  DrawText(Canvas.Handle, PChar(FMessage), -1, Rect, DT_CENTER or DT_NOPREFIX);
  SetBkMode(Canvas.Handle, OldBkMode);
end;

procedure TWaitForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := FThread = nil;
end;

procedure TWaitForm.FormFirstShow(Sender: TObject);
begin
  if FThread <> nil then FThread.Resume
  else Close;
end;

procedure TWaitForm.UMFirstShow(var Message: TMessage);
begin
  FormFirstShow(Self);
end;

procedure TWaitForm.FormShow(Sender: TObject);
begin
  if not ShowedOnce then
  begin
    ShowedOnce := True;
    PostMessage(Handle, UM_FIRST_SHOW, 0, 0);
  end;
end;

constructor TWaitForm.Create(AOwner: TComponent);
begin
  inherited;
  ShowedOnce := False;
end;

procedure TWaitForm.ApplicationEventsIdle(Sender: TObject;
  var Done: Boolean);
var
  ExitCode: LongWord;
begin
  if not GetExitCodeThread(FThread.Handle, ExitCode) or (ExitCode <> STILL_ACTIVE) then
  begin
    FThread := nil;
    Close;
  end;
end;

end.

