unit VFImageFormatTIFF;

interface

uses SysUtils, Classes, VFImage;

type
  TVFImageFormatTIFF = class(TVFImageFormat)
  protected
    function GetCanLoad: Boolean; override;
    function GetCanSave: Boolean; override;
    function GetFileFilter: string; override;
    function GetFileFilterName: string; override;

    procedure Load(Image: TVFImage; Stream: TStream); override;

  end;

implementation

var
  ImageFormatTIFF: TVFImageFormatTIFF;

const
  TIFF_TAG_NewSubfileType = 254;
  TIFF_TAG_SubfileType = 255;
  TIFF_TAG_ImageWidth = 256;
  TIFF_TAG_ImageLength = 257;
  TIFF_TAG_BitsPerSample = 258;
  TIFF_TAG_Compression = 259;
  TIFF_TAG_VALUE_Uncompressed = 1;
  TIFF_TAG_VALUE_CCITT_1D = 2;
  TIFF_TAG_VALUE_Group_3_FAX = 3;
  TIFF_TAG_VALUE_Group_4_FAX = 4;
  TIFF_TAG_VALUE_LZW = 5;
  TIFF_TAG_VALUE_JPEG = 6;
  TIFF_TAG_VALUE_PackBits = 32773;
  TIFF_TAG_PhotometricInterpretation = 262;
  TIFF_TAG_VALUE_WhiteIsZero = 0;
  TIFF_TAG_VALUE_BlackIsZero = 1;
  TIFF_TAG_VALUE_RGB = 2;
  TIFF_TAG_VALUE_RGB_Palette = 3;
  TIFF_TAG_VALUE_Transparency_mask = 4;
  TIFF_TAG_VALUE_CMYK = 5;
  TIFF_TAG_VALUE_YCbCr = 6;
  TIFF_TAG_VALUE_CIELab = 8;
  TIFF_TAG_Threshholding = 263;
  TIFF_TAG_CellWidth = 264;
  TIFF_TAG_CellLength = 265;
  TIFF_TAG_FillOrder = 266;
  TIFF_TAG_DocumentName = 269;
  TIFF_TAG_ImageDescription = 270;
  TIFF_TAG_Make = 271;
  TIFF_TAG_Model = 272;
  TIFF_TAG_StripOffsets = 273;
  TIFF_TAG_Orientation = 274;
  TIFF_TAG_SamplesPerPixel = 277;
  TIFF_TAG_RowsPerStrip = 278;
  TIFF_TAG_StripByteCounts = 279;
  TIFF_TAG_MinSampleValue = 280;
  TIFF_TAG_MaxSampleValue = 281;
  TIFF_TAG_XResolution = 282;
  TIFF_TAG_YResolution = 283;
  TIFF_TAG_PlanarConfiguration = 284;
  TIFF_TAG_PageName = 285;
  TIFF_TAG_XPosition = 286;
  TIFF_TAG_YPosition = 287;
  TIFF_TAG_FreeOffsets = 288;
  TIFF_TAG_FreeByteCounts = 289;
  TIFF_TAG_GrayResponseUnit = 290;
  TIFF_TAG_GrayResponseCurve = 291;
  TIFF_TAG_T4Options = 292;
  TIFF_TAG_T6Options = 293;
  TIFF_TAG_ResolutionUnit = 296;
  TIFF_TAG_PageNumber = 297;
  TIFF_TAG_TransferFunction = 301;
  TIFF_TAG_Software = 305;
  TIFF_TAG_DateTime = 306;
  TIFF_TAG_Artist = 315;
  TIFF_TAG_HostComputer = 316;
  TIFF_TAG_Predictor = 317;
  TIFF_TAG_WhitePoint = 318;
  TIFF_TAG_PrimaryChromaticities = 319;
  TIFF_TAG_ColorMap = 320;
  TIFF_TAG_HalftoneHints = 321;
  TIFF_TAG_TileWidth = 322;
  TIFF_TAG_TileLength = 323;
  TIFF_TAG_TileOffsets = 324;
  TIFF_TAG_TileByteCounts = 325;
  TIFF_TAG_InkSet = 332;
  TIFF_TAG_InkNames = 333;
  TIFF_TAG_NumberOfInks = 334;
  TIFF_TAG_DotRange = 336;
  TIFF_TAG_TargetPrinter = 337;
  TIFF_TAG_ExtraSamples = 338;
  TIFF_TAG_SampleFormat = 339;
  TIFF_TAG_SMinSampleValue = 340;
  TIFF_TAG_SMaxSampleValue = 341;
  TIFF_TAG_TransferRange = 342;
  TIFF_TAG_JPEGProc = 512;
  TIFF_TAG_JPEGInterchangeFormat = 513;
  TIFF_TAG_JPEGInterchangeFormatLngth = 514;
  TIFF_TAG_JPEGRestartInterval = 515;
  TIFF_TAG_JPEGLosslessPredictors = 517;
  TIFF_TAG_JPEGPointTransforms = 518;
  TIFF_TAG_JPEGQTables = 519;
  TIFF_TAG_JPEGDCTables = 520;
  TIFF_TAG_JPEGACTables = 521;
  TIFF_TAG_YCbCrCoefficients = 529;
  TIFF_TAG_YCbCrSubSampling = 530;
  TIFF_TAG_YCbCrPositioning = 531;
  TIFF_TAG_ReferenceBlackWhite = 532;
  TIFF_TAG_Copyright = 33432;

  TIFF_TYPE_Byte   = 1;
  TIFF_TYPE_Short  = 3;
  TIFF_TYPE_Long   = 4;
  TIFF_TYPE_SByte  = 6;
  TIFF_TYPE_SShort = 8;
  TIFF_TYPE_SLong  = 9;

  TIFF_LZW_EoiCode       = 257;
  TIFF_LZW_ClearCode     = 256;
  TIFF_LZW_MaxTableEntry = 8193;

type
  PTag = ^TTag;
  TTag = record
    Tag: Word;
    ValueType: Word;
    Count: LongWord;
    Value: record
      case Integer of
        0: (Byte: array[0..3] of Byte);
        1: (Short: array[0..1] of Word);
        2: (Long: LongWord);
      end;
  end;

  TLZWTableEntry = record
    Length: Word;
    Value: Byte;
    OldCode: Word;
  end;

  TByteArray = array of Byte;

  TLongWordArray = array of LongWord;

  TTagArray = array of TTag;

  TFillOrderTable = array[0..254] of Byte;

  TLZWTable = array[0..TIFF_LZW_MaxTableEntry - 1] of TLZWTableEntry;

function TIFFGetValueByOrder(Value: Word; ByteOrder: Boolean): Word; overload;
begin
  if ByteOrder then Result := Value
  else Result := Word((Value shl 8) or (value shr 8));
end;

function TIFFGetValueByOrder(Value: LongWord; ByteOrder: Boolean): LongWord; overload;
begin
  if ByteOrder then Result := Value
  else Result := LongWord((Value shl 24) or ((Value and $FF00) shl 8)
      or ((Value and $FF0000) shr 8) or ((Value and $FF000000) shr 24));
end;

function TIFFGetTag(TagsCount: Word; Tags: TTagArray; TagType: Word; ByteOrder: Boolean): PTag;
var
  I: Word;
begin
  for I := 0 to TagsCount - 1 do
    if TagType = TIFFGetValueByOrder(Tags[i].Tag, ByteOrder) then
    begin
      Result := @Tags[I];
      Exit;
    end;
  Result := nil;
end;

function TIFFGetTagValue(TagsCount: Word; Tags: TTagArray; TagType: Word; ByteOrder: Boolean): LongWord;
var
  T: PTag;
begin
  T := TIFFGetTag(TagsCount, Tags, TagType, ByteOrder);
  if T <> nil then
    if TIFFGetValueByOrder(T.Count, ByteOrder) = 1 then
      case TIFFGetValueByOrder(T.ValueType, ByteOrder) of
        TIFF_TYPE_Byte, TIFF_TYPE_SByte: begin
          Result := LongWord(T.Value.Byte[0]);
          Exit;
        end;

        TIFF_TYPE_Short, TIFF_TYPE_SShort: begin
          Result := LongWord(TIFFGetValueByOrder(T.Value.Short[0], ByteOrder));
          Exit;
        end;

        TIFF_TYPE_Long, TIFF_TYPE_SLong: begin
          Result := LongWord(TIFFGetValueByOrder(T.Value.Long, ByteOrder));
          Exit;
        end;
      end;
  case TagType of
    TIFF_TAG_BitsPerSample: Result := 1;
    TIFF_TAG_Compression:  Result := 1;
    TIFF_TAG_FillOrder: Result := 1;
    TIFF_TAG_MaxSampleValue: Result := 0; //Default is 2**(BitsPerSample) - 1.
    TIFF_TAG_MinSampleValue: Result := 0;
    TIFF_TAG_NewSubfileType: Result := 0;
    TIFF_TAG_Orientation: Result := 1;
    TIFF_TAG_PlanarConfiguration: Result := 1;
    TIFF_TAG_ResolutionUnit: Result := 2;
    TIFF_TAG_SamplesPerPixel: Result := 1;
    TIFF_TAG_Threshholding: Result := 1;
  else
    Result := 0;
  end;
end;

procedure TIFFGetTagValues(Stream: TStream; TagsCount: Word; Tags: TTagArray;
  TagType: Word; var TagValues: TLongWordArray; ByteOrder: Boolean);
var
  T: PTag;
  I: LongWord;
  TempByte: Byte;
  TempShort: Word;
  TempLong: LongWord;
begin
  SetLength(TagValues, 0);
  T := TIFFGetTag(TagsCount, Tags, TagType, ByteOrder);
  if T <> nil then
  begin
    if TIFFGetValueByOrder(T.Count, ByteOrder) = 0 then Exit;
    SetLength(TagValues, TIFFGetValueByOrder(T.Count, ByteOrder));
    case TIFFGetValueByOrder(T.ValueType, ByteOrder) of
      TIFF_TYPE_Byte, TIFF_TYPE_SByte:
        if TIFFGetValueByOrder(T.Count, ByteOrder) <= 4 then
        begin
          for I := 0 to TIFFGetValueByOrder(T.Count, ByteOrder) - 1 do
            TagValues[I] := LongWord(T.Value.Byte[I]);
        end
        else begin
          Stream.Seek(TIFFGetValueByOrder(T.Value.Long, ByteOrder), soFromBeginning);
          for I := 0 to TIFFGetValueByOrder(T.Count, ByteOrder) - 1 do
          begin
            Stream.ReadBuffer(TempByte, SizeOf(Byte));
            TagValues[I] := TempByte;
          end;
        end;
      TIFF_TYPE_Short, TIFF_TYPE_SShort:
        if TIFFGetValueByOrder(T.Count, ByteOrder) <= 2 then
        begin
          for I := 0 to TIFFGetValueByOrder(T.Count, ByteOrder) - 1 do
            TagValues[I] := LongWord(TIFFGetValueByOrder(T.Value.Short[I], ByteOrder));
        end
        else begin
          Stream.Seek(TIFFGetValueByOrder(T.Value.Long, ByteOrder), soFromBeginning);
          for I := 0 to TIFFGetValueByOrder(T.Count, ByteOrder) - 1 do
          begin
            Stream.ReadBuffer(TempShort, SizeOf(Word));
            TagValues[I] := TIFFGetValueByOrder(TempShort, ByteOrder);
          end;
        end;
      TIFF_TYPE_Long, TIFF_TYPE_SLong:
        if TIFFGetValueByOrder(T.Count, ByteOrder) <= 1 then
        begin
          TagValues[0] := TIFFGetValueByOrder(T.Value.Long, ByteOrder);
        end
        else begin
          Stream.Seek(TIFFGetValueByOrder(T.Value.Long, ByteOrder), soFromBeginning);
          for I := 0 to TIFFGetValueByOrder(T.Count, ByteOrder) - 1 do
          begin
            Stream.ReadBuffer(TempLong, SizeOf(LongWord));
            TagValues[I] := TIFFGetValueByOrder(TempLong, ByteOrder);
          end;
        end;
    else
      SetLength(TagValues, 0);
    end;
  end;
end;

function GetBit(Value, Bit: Byte): Byte;
begin
  Result := (Value shr (7 - Bit)) and $01;
end;

function Get4Bit(Value: Byte; Part: Boolean): Byte;
begin
  if Part then Result := Value and $0F
  else Result := (Value shr 4) and $0F;
end;

procedure FillOrderTable(var Table: TFillOrderTable);
var
  I: Byte;
begin
  for I := 0 to 254 do
    Table[I] := ((I shr 7) and $01)
      or ((I shr 6) and $01 shl 1)
      or ((I shr 5) and $01 shl 2)
      or ((I shr 4) and $01 shl 3)
      or ((I shr 3) and $01 shl 4)
      or ((I shr 2) and $01 shl 5)
      or ((I shr 1) and $01 shl 6)
      or ((I)       and $01 shl 7);
end;

procedure ChangeFillOrder(var Table: TFillOrderTable; Data: TByteArray; L: LongWord);
var
  I: LongWord;
begin
  for I := 0 to L - 1 do
    Data[I] := Table[Data[I]];
end;

function TIFFLZWGetNextCode(DataIn: TByteArray; var StreamByte, StreamBit, StreamCodeBits: LongWord): Word;
var
  I: LongWord;
begin
  Result := 0;
  for I := 0 to StreamCodeBits - 1 do
  begin
    Result := Result shl 1;
    if DataIn[StreamByte] and $80 <> 0 then Inc(Result);
    DataIn[StreamByte] := DataIn[StreamByte] shl 1;
    Inc(StreamBit);
    if StreamBit > 7 then
    begin
      Inc(StreamByte);
      StreamBit := 0;
    end;
  end;
end;

procedure TIFFLZWInitializeTable(var Table: TLZWTable);
var
  I: Word;
begin
  for I := 256 to TIFF_LZW_MaxTableEntry - 1 do
  begin
    Table[I].Length := 0;
    Table[I].OldCode := TIFF_LZW_EoiCode;
  end;
end;

procedure TIFFLZWWriteCode(DataOut: TByteArray; var Writed: LongWord; var Table: TLZWTable; Code: Word; var FirstChar: Byte; Length: LongWord);
var
  I, L: LongWord;
begin
  L := Table[Code].Length;
  for I := 0 to L - 1 do
  begin
    DataOut[Writed + L - I - 1] := Table[Code].Value;
    if Table[Code].OldCode <> TIFF_LZW_EoiCode then Code := Table[Code].OldCode
    else Break;
  end;
  FirstChar := Table[Code].Value;
  Inc(Writed, L);
end;

function TIFFLZWIsInTable(var Table: TLZWTable; Code: Word): Boolean;
begin
  Result := Table[Code].Length > 0;
end;

procedure TIFFLZWAddStringToTable(var Table: TLZWTable; Code1: Word; FirstChar: Byte; var StreamCodeBits: LongWord; Entry: Word);
var
  I: Word;
begin
  for I := Entry to TIFF_LZW_MaxTableEntry - 1 do
  begin
    if Table[I].Length = 0 then Break;
  end;
  Table[I].Length := Table[Code1].Length + 1;
  Table[I].OldCode := Code1;
  Table[I].Value := FirstChar;
  StreamCodeBits := 9;
  if I >= 511-1 then StreamCodeBits := 10;
  if I >= 1023-1 then streamCodeBits := 11;
  if I >= 2047-1 then streamCodeBits := 12;
end;

procedure TIFFDecompress(DataIn, DataOut: TByteArray; L, BytesPerRow, Compression: LongWord);
var
  I: Word;
  Writed, Readed, R: LongWord;
  Value: ShortInt;
  Code, OldCode: Word;
  StreamByte, StreamBit, StreamCodeBits: LongWord;
  FirstChar: Byte;
  Table: TLZWTable;
  EntryCount: Word;
begin
  for I := 0 to TIFF_LZW_MaxTableEntry - 1 do
  begin
    Table[I].Length := 0;
    Table[I].OldCode := TIFF_LZW_EoiCode;
  end;
  case Compression of
    TIFF_TAG_VALUE_Uncompressed: Move(DataIn[0], DataOut[0], L);
    TIFF_TAG_VALUE_PackBits: begin
      Writed := 0;
      Readed := 0;
      R := 0;
      while Writed < L do
      begin
        Writed := R * BytesPerRow;
        while Writed < (R + 1) * BytesPerRow do
        begin
          Value := ShortInt(DataIn[Readed]); Inc(Readed);
          if Value >= 0 then
          begin
            for I := 0 to Value do
            begin
              DataOut[Writed] := DataIn[Readed];
              Inc(Readed); Inc(Writed);
              if Writed >= (R + 1) * BytesPerRow then Break;
            end;
          end;
          if Writed >= (R + 1) * BytesPerRow then Break;
          if (Value >= -127) and (Value <= -1) then
          begin
            for I := 0 to -Value do
            begin
              DataOut[Writed] := DataIn[Readed]; Inc(Writed);
              if Writed >= (R + 1) * BytesPerRow then Break;
            end;
            Inc(Readed);
          end;
        end;
        Inc(R);
      end;
    end;
  TIFF_TAG_VALUE_LZW: begin
    StreamCodeBits := 9;
    OldCode := 0;
    StreamByte := 0;
    StreamBit := 0;
    EntryCount := 258;
    for I := 0 to 255 do
    begin
      Table[I].Length := 1;
      Table[I].Value := Byte(I);
    end;
    Writed := 0;
    repeat begin
      if Writed >= L then Break;
      Code := TIFFLZWGetNextCode(DataIn, StreamByte, StreamBit, StreamCodeBits);
      if Code = TIFF_LZW_EoiCode then Break;
      if (Code = TIFF_LZW_ClearCode) or (Writed = 0) then
      begin
        TIFFLZWInitializeTable(Table); EntryCount := 258; StreamCodeBits := 9;
        Code := TIFFLZWGetNextCode(DataIn, StreamByte, StreamBit, StreamCodeBits);
        if Code = TIFF_LZW_EoiCode then Break;
        TIFFLZWWriteCode(DataOut, Writed, Table, Code, FirstChar, L);
        OldCode := Code;
      end
      else begin
        if TIFFLZWIsInTable(Table, Code) then
        begin
          TIFFLZWWriteCode(DataOut, Writed, Table, Code, FirstChar, L);
          TIFFLZWAddStringToTable(Table, OldCode, FirstChar, StreamCodeBits, EntryCount);
          Inc(EntryCount);
          OldCode := Code;
        end
        else begin
          TIFFLZWWriteCode(DataOut, Writed, Table, OldCode, FirstChar, L);
          DataOut[Writed] := FirstChar; Inc(Writed);
          TIFFLZWAddStringToTable(Table, OldCode, FirstChar, StreamCodeBits, Code);
          EntryCount := Code + 1;
          OldCode := Code;
        end;
      end;
    end until False;
  end;
  else
    raise Exception.Create('Unknown compression');
  end;
end;

procedure TIFFReadImage(Stream: TStream; Image: TVFImage; TagsCount: Word; Tags: TTagArray;
  BitsPerSample: LongWord; RGB: Boolean; ColorMap: TLongWordArray; ByteOrder: Boolean);
var
  TagPhoto, TagFillOrder, TagCompression, TagPredictor: LongWord;
  Table: TFillOrderTable;
  T: PTag;
  Tiles: Boolean;
  Count: LongWord;
  Offsets, Counts: TLongWordArray;
  TileWidth, TileLength, BytesPerRow: LongWord;
  TilePerX: LongWord;
  MaxTile: LongWord;
  I, J, Y, X, Z, MaxX, MaxY: LongWord;
  Data, UncompressedData: TByteArray;
  Temp: LongWord;
  RGBSize: LongWord;
  RowsPerStrip: LongWord;
  MaxStrip: LongWord;
begin
  try
    TagPredictor := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_Predictor, ByteOrder);
    TagPhoto := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_PhotometricInterpretation, ByteOrder);
    TagFillOrder := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_FillOrder, ByteOrder);
    TagCompression := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_Compression, ByteOrder);
    if TagFillOrder = 2 then FillOrderTable(Table);

    TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_StripOffsets, Offsets, ByteOrder);
    TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_StripByteCounts, Counts, ByteOrder);
    T := TIFFGetTag(TagsCount, Tags, TIFF_TAG_StripByteCounts, ByteOrder);
    if T = nil then Count := 0 else Count := TIFFGetValueByOrder(T.Count, ByteOrder);
    if (Length(Offsets) = 0) or (Length(Counts) = 0) or (Count = 0) then
    begin
      TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_TileOffsets, Offsets, ByteOrder);
      TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_TileByteCounts, Counts, ByteOrder);
      T := TIFFGetTag(TagsCount, Tags, TIFF_TAG_TileOffsets, ByteOrder);
      if T = nil then Count := 0 else Count := TIFFGetValueByOrder(T.Count, ByteOrder);
      if (Length(Offsets) = 0) or (Length(Counts) = 0) or (Count = 0) then
        raise Exception.Create('Invalid image')
      else Tiles := true;
    end
    else Tiles := false;

    if Tiles then
    begin
      TileWidth := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_TileWidth, ByteOrder);
      TileLength := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_TileLength, ByteOrder);
      BytesPerRow := 0;
      if BitsPerSample = 1 then BytesPerRow := (TileWidth + 7) div 8
      else if BitsPerSample = 4 then BytesPerRow := (TileWidth + 1) div 2
      else if BitsPerSample = 8 then BytesPerRow := TileWidth
      else if BitsPerSample = 16 then BytesPerRow := TileWidth * 2;
      if RGB then BytesPerRow := BytesPerRow * 3;
      TilePerX := LongWord(Image.Width) div TileWidth; if LongWord(Image.Width) mod TileWidth <> 0 then Inc(TilePerX);
      MaxTile := Counts[0];
      for I := 1 to Count - 1 do
        if MaxTile < Counts[I] then MaxTile := Counts[I];
      SetLength(Data, MaxTile);
      SetLength(UncompressedData, TileLength * BytesPerRow);
      for I := 0 to Count - 1 do
      begin
        Stream.Position := Offsets[I];
        Stream.ReadBuffer(Data[0], Counts[I]);
        if TagFillOrder = 2 then ChangeFillOrder(Table, Data, Counts[I]);
        TIFFDecompress(Data, UncompressedData, BytesPerRow * TileLength, BytesPerRow, TagCompression);
        J := I div TilePerX * TileLength;
        Z := I mod TilePerX * TileWidth;
        if J + TileLength > LongWord(Image.Height) then MaxY := LongWord(Image.Height)
        else MaxY := LongWord(J) + TileLength;
        if Z + TileWidth > LongWord(Image.Width) then MaxX := LongWord(Image.Width)
        else MaxX := Z + TileWidth;
        if RGB then
        begin
          if BitsPerSample = 8 then
          begin
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
                begin
                  if (TagPredictor = 2) and (X > Z) then // Horizontal differencing
                  begin
                    UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3    ] := UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3    ] + UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3    ];
                    UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 1] := UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 1] + UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3 + 1];
                    UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 2] := UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 2] + UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3 + 2];
                  end;
                  Image.Pixels[X, Y] :=
                    ( UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3    ]
                    + UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 1]
                    + UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 + 2] ) div 3;
                end;
          end
          else if BitsPerSample = 16 then
          begin
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
              begin
                if (TagPredictor = 2) and (X > Z) then // Horizontal differencing
                begin
                  PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2    ])^ := PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2    ])^ + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3 * 2    ])^;
                  PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 1])^ := PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 1])^ + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3 * 2 + 1])^;
                  PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 2])^ := PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 2])^ + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 3 * 2 + 2])^;
                end;
                Image.Pixels[X, Y] := Byte(
                  ( PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2    ])^
                  + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 1])^
                  + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 3 * 2 + 2])^ ) div 3 div $100);
              end;
          end;
        end
        else if BitsPerSample = 1 then begin
          for Y := J to MaxY - 1 do
            for X := Z to MaxX - 1 do
              Image.Pixels[X, Y] := GetBit(UncompressedData[(Y - J) * BytesPerRow + (X - Z) div 8], Byte((X - Z) mod 8)) * 255;
        end
        else if Length(ColorMap) = 0 then begin
          if BitsPerSample = 4 then
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
                Image.Pixels[X, Y] := Get4Bit(UncompressedData[(Y - J) * BytesPerRow + (X - Z) div 2], (X - Z) mod 2 <> 0) * 15
          else if BitsPerSample = 8 then
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
              begin
                if (TagPredictor = 2) and (X > Z) then // Horizontal differencing
                  UncompressedData[(Y - J) * BytesPerRow + (X - Z)] := UncompressedData[(Y - J) * BytesPerRow + (X - Z)] + UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1)];
                Image.Pixels[X, Y] := UncompressedData[(Y - J) * BytesPerRow + (X - Z)];
              end
          else if BitsPerSample = 16 then
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
              begin
                if (TagPredictor = 2) and (X > Z) then // Horizontal differencing
                  PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 2])^ := PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 2])^ + PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1) * 2])^;
                Image.Pixels[X, Y] := Byte(PWord(@UncompressedData[(Y - J) * BytesPerRow + (X - Z) * 2])^ div $100);
              end;
        end
        else begin
          RGBSize := 1 shl BitsPerSample;
          if BitsPerSample = 4 then
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
              begin
                Temp := Get4Bit(UncompressedData[(Y - J) * BytesPerRow + (X - Z) div 2], (X - Z) mod 2 <> 0);
                Image.Pixels[X, Y] := Byte((ColorMap[Temp] + ColorMap[RGBSize + Temp] + ColorMap[2 * RGBSize + Temp]) div 3 div $100);
              end
          else if BitsPerSample = 8 then
            for Y := J to MaxY - 1 do
              for X := Z to MaxX - 1 do
              begin
                if (TagPredictor = 2) and (X > Z) then // Horizontal differencing
                  UncompressedData[(Y - J) * BytesPerRow + (X - Z)] := UncompressedData[(Y - J) * BytesPerRow + (X - Z)] + UncompressedData[(Y - J) * BytesPerRow + (X - Z - 1)];
                Temp := UncompressedData[(Y - J) * BytesPerRow + (X - Z)];
                Image.Pixels[X, Y] := Byte((ColorMap[Temp] + ColorMap[RGBSize + Temp] + ColorMap[2 * RGBSize + Temp]) div 3 div $100);
              end;
        end;
      end;
    end
    else begin
      RowsPerStrip := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_RowsPerStrip, ByteOrder);
      BytesPerRow := 0;
      if BitsPerSample = 1 then BytesPerRow := (Image.Width + 7) div 8
      else if BitsPerSample = 4 then BytesPerRow := (Image.Width + 1) div 2
      else if BitsPerSample = 8 then BytesPerRow := Image.Width
      else if BitsPerSample = 16 then BytesPerRow := Image.Width * 2;
      if RGB then BytesPerRow := BytesPerRow * 3;
      if RowsPerStrip = 0 then RowsPerStrip := Counts[0] div BytesPerRow;
      MaxStrip := Counts[0];
      for I := 1 to Count - 1 do
        if MaxStrip < Counts[I] then MaxStrip := Counts[i];
      SetLength(Data, MaxStrip);
      SetLength(UncompressedData, RowsPerStrip * BytesPerRow);
      for I := 0 to Count - 1 do
      begin
        Stream.Position := Offsets[I];
        Stream.ReadBuffer(Data[0], Counts[I]);
        if TagFillOrder = 2 then ChangeFillOrder(Table, Data, Counts[I]);
        if I = Count - 1 then J := LongWord(Image.Height) - RowsPerStrip * I
        else J := RowsPerStrip;
        TIFFDecompress(Data, UncompressedData, BytesPerRow * j, BytesPerRow, TagCompression);
        if RGB then
        begin
          if BitsPerSample = 8 then
          begin
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
              begin
                if (TagPredictor = 2) and (X > 0) then // Horizontal differencing
                begin
                  UncompressedData[Y * BytesPerRow + X * 3    ] := UncompressedData[Y * BytesPerRow + X * 3    ] + UncompressedData[Y * BytesPerRow + (X - 1) * 3    ];
                  UncompressedData[Y * BytesPerRow + X * 3 + 1] := UncompressedData[Y * BytesPerRow + X * 3 + 1] + UncompressedData[Y * BytesPerRow + (X - 1) * 3 + 1];
                  UncompressedData[Y * BytesPerRow + X * 3 + 2] := UncompressedData[Y * BytesPerRow + X * 3 + 2] + UncompressedData[Y * BytesPerRow + (X - 1) * 3 + 2];
                end;
                Image.Pixels[X, RowsPerStrip * I + Y] :=
                  ( UncompressedData[Y * BytesPerRow + X * 3    ]
                  + UncompressedData[Y * BytesPerRow + X * 3 + 1]
                  + UncompressedData[Y * BytesPerRow + X * 3 + 2] ) div 3;
              end;
          end
          else if BitsPerSample = 16 then
          begin
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
              begin
                if (TagPredictor = 2) and (X > 0) then // Horizontal differencing
                begin
                  PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2    ])^ := PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2    ])^ + PWord(@UncompressedData[Y * BytesPerRow + (X - 1) * 3 * 2    ])^;
                  PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2 + 1])^ := PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2 + 1])^ + PWord(@UncompressedData[Y * BytesPerRow + (X - 1) * 3 * 2 + 1])^;
                  PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2 + 2])^ := PWord(@UncompressedData[Y * BytesPerRow + X * 3 * 2 + 2])^ + PWord(@UncompressedData[Y * BytesPerRow + (X - 1) * 3 * 2 + 2])^;
                end;
                Image.Pixels[X, RowsPerStrip * I + Y] := Byte(
                  ( PWord(@UncompressedData[Y * BytesPerRow + X * 2 * 3    ])^
                  + PWord(@UncompressedData[Y * BytesPerRow + X * 2 * 3 + 1])^
                  + PWord(@UncompressedData[Y * BytesPerRow + X * 2 * 3 + 2])^ ) div 3 div $100);
              end;
          end;
        end
        else if BitsPerSample = 1 then
        begin
          for Y := 0 to J - 1 do
            for X := 0 to Image.Width - 1 do
              Image.Pixels[X, RowsPerStrip * I + Y] := GetBit(UncompressedData[Y * BytesPerRow + X div 8], Byte(X mod 8)) * 255;
        end
        else if Length(ColorMap) = 0 then
        begin
          if BitsPerSample = 4 then
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
                Image.Pixels[X, RowsPerStrip * I + Y] := Get4Bit(UncompressedData[Y * BytesPerRow + X div 2], X mod 2 <> 0) * 15
          else if BitsPerSample = 8 then
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
              begin
                if (TagPredictor = 2) and (X > 0) then // Horizontal differencing
                  UncompressedData[Y * BytesPerRow + X] := UncompressedData[Y * BytesPerRow + X] + UncompressedData[Y * BytesPerRow + X - 1];
                Image.Pixels[X, RowsPerStrip * I + Y] := UncompressedData[Y * BytesPerRow + X];
              end
          else if BitsPerSample = 16 then
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
                begin
                  if (TagPredictor = 2) and (X > 0) then // Horizontal differencing
                    PWord(@UncompressedData[Y * BytesPerRow + 2 * (X - 1)])^ := PWord(@UncompressedData[Y * BytesPerRow + 2 * (X - 1)])^ + PWord(@UncompressedData[Y * BytesPerRow + 2 * (X - 1)])^;
                  Image.Pixels[X, RowsPerStrip * I + Y] := Byte(PWord(@UncompressedData[Y * BytesPerRow + 2 * x])^ div $100);
                end;
        end
        else begin
          RGBSize := 1 shl BitsPerSample;
          if BitsPerSample = 4 then
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
              begin
                Temp := Get4Bit(UncompressedData[Y * BytesPerRow + X div 2], X mod 2 <> 0);
                Image.Pixels[X, RowsPerStrip * I + Y] := Byte((ColorMap[Temp] + ColorMap[RGBSize + Temp] + ColorMap[2 * RGBSize + Temp]) div 3 div $100);
              end
          else if BitsPerSample = 8 then
            for Y := 0 to J - 1 do
              for X := 0 to Image.Width - 1 do
              begin
                if (TagPredictor = 2) and (X > 0) then // Horizontal differencing
                  UncompressedData[Y * BytesPerRow + X] := UncompressedData[Y * BytesPerRow + X] + UncompressedData[Y * BytesPerRow + X - 1];
                Temp := UncompressedData[Y * BytesPerRow + X];
                Image.Pixels[X, RowsPerStrip * I + Y] := Byte((ColorMap[Temp] + ColorMap[RGBSize + Temp] + ColorMap[2 * RGBSize + Temp]) div 3 div $100);
              end;
        end;
      end;
    end;
    if TagPhoto = TIFF_TAG_VALUE_WhiteIsZero then
      for I := 0 to Image.Height - 1 do
        for J := 0 to Image.Width - 1 do
          Image.Pixels[J, I] := 255 - Image.Pixels[J, I];
  finally
    SetLength(Offsets, 0);
    SetLength(Counts, 0);
    SetLength(Data, 0);
    SetLength(UncompressedData, 0);
  end;
end;

{ TVFImageFormatTIFF }

function TVFImageFormatTIFF.GetCanLoad: Boolean;
begin
  Result := True;
end;

function TVFImageFormatTIFF.GetCanSave: Boolean;
begin
  Result := False;
end;

function TVFImageFormatTIFF.GetFileFilter: string;
begin
  Result := '*.tif';
end;

function TVFImageFormatTIFF.GetFileFilterName: string;
begin
  Result := 'TIFF files';
end;

procedure TVFImageFormatTIFF.Load(Image: TVFImage; Stream: TStream);
var
  Signature: array[0..1] of AnsiChar;
  IDFAddress: LongWord;
  I: Word;
  W, H: Word;
  ByteOrder: Boolean;
  TagsCount: Word;
  Tags: TTagArray;
  ColorMap, BitPerSample: TLongWordArray;
  BitsPerSample, SamplesPerPixel, PhotometricInterpretation: LongWord;
begin
  try
    Stream.ReadBuffer(Signature, SizeOf(Signature));
    if (Signature[1] <> Signature[0]) and ((Signature[0] <> 'I') or (Signature[0] <> 'M')) then
      raise Exception.Create('Invalid TIFF signature');
    ByteOrder := Signature[0] = 'I';
    Stream.Seek(2, soFromCurrent);
    Stream.ReadBuffer(IDFAddress, SizeOf(LongWord));
    Stream.Seek(TIFFGetValueByOrder(IDFAddress, ByteOrder), soFromBeginning);
    Stream.ReadBuffer(TagsCount, SizeOf(Word));
    TagsCount := TIFFGetValueByOrder(TagsCount, ByteOrder);
    if TagsCount = 0 then raise Exception.Create('Invalid tags count');
    SetLength(Tags, TagsCount);
    for I := 0 to TagsCount - 1 do
      Stream.ReadBuffer(Tags[I], SizeOf(TTag));
    TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_ColorMap, ColorMap, ByteOrder);
    TIFFGetTagValues(Stream, TagsCount, Tags, TIFF_TAG_BitsPerSample, BitPerSample, ByteOrder);
    W := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_ImageWidth, ByteOrder);
    H := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_ImageLength, ByteOrder);
    if (W = 0) or (H = 0) then raise Exception.Create('Invalid image size');
    Image.SetSize(W, H);
    BitsPerSample := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_BitsPerSample, ByteOrder);
    SamplesPerPixel := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_SamplesPerPixel, ByteOrder);
    PhotometricInterpretation := TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_PhotometricInterpretation, ByteOrder);
    if (BitsPerSample = 1) and (SamplesPerPixel = 1)
      and ((PhotometricInterpretation = TIFF_TAG_VALUE_WhiteIsZero) or (PhotometricInterpretation = TIFF_TAG_VALUE_BlackIsZero)) then
      TIFFReadImage(Stream, Image, TagsCount, Tags, BitsPerSample, False, ColorMap, ByteOrder)
    else if ((BitsPerSample = 4) or (BitsPerSample = 8) or (BitsPerSample = 16)) and (SamplesPerPixel = 1)
      and ((PhotometricInterpretation = TIFF_TAG_VALUE_WhiteIsZero) or (PhotometricInterpretation = TIFF_TAG_VALUE_BlackIsZero)) then
      TIFFReadImage(Stream, Image, TagsCount, Tags, BitsPerSample, False, ColorMap, ByteOrder)
    else if ((BitsPerSample = 4) or (BitsPerSample = 8)) and (Length(ColorMap) <> 0)
      and (TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_PhotometricInterpretation, ByteOrder) = TIFF_TAG_VALUE_RGB_Palette) then
      TIFFReadImage(Stream, Image, TagsCount, Tags, BitsPerSample, False, ColorMap, ByteOrder)
    else if (Length(BitPerSample) <> 0)
      and (TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_SamplesPerPixel, ByteOrder) = 3)
      and (TIFFGetTagValue(TagsCount, Tags, TIFF_TAG_PhotometricInterpretation, ByteOrder) = TIFF_TAG_VALUE_RGB) then
    begin
      if (BitPerSample[0] = 8) and (BitPerSample[1] = 8) and (BitPerSample[2] = 8) then
        TIFFReadImage(Stream, Image, TagsCount, Tags, 8, True, ColorMap, ByteOrder)
      else if (BitPerSample[0] = 16) and (BitPerSample[1] = 16) and (BitPerSample[2] = 16) then
        TIFFReadImage(Stream, Image, TagsCount, Tags, 16, True, ColorMap, ByteOrder);
    end
    else raise Exception.Create('Invalid image format');
  finally
    SetLength(Tags, 0);
    SetLength(ColorMap, 0);
    SetLength(BitPerSample, 0);
  end;
end;

initialization
  ImageFormatTIFF := TVFImageFormatTIFF.Create;

finalization
  ImageFormatTIFF.Free;

end.

