//#if defined(linux) || defined (__LINUX__)
//	#include "linux_include.h"
//#else
	#include "stdafx.h"
//#endif

bool SetIntArrayField(JNIEnv * env, jclass* cls, jobject* object, char* name, int* data, int size)
{
    jfieldID fid = env->GetFieldID(*cls, name,   "[I");
    if (fid == NULL)
        return false;

    jintArray jarray = (jintArray) env->GetObjectField(*object, fid);
    if (jarray == NULL)
        return false;

    jint* array = env->GetIntArrayElements(jarray, NULL);
    if (array == NULL)
    {
        env->DeleteLocalRef(jarray);
        return false;
    }

    memcpy(array, data, size*sizeof(int));

    env->ReleaseIntArrayElements(jarray, array, 0);

    env->DeleteLocalRef(jarray);

    return true;
}

bool SetByteArrayField(JNIEnv * env, jclass* cls, jobject* object, char* name, BYTE* data, int size)
{
    jfieldID fid = env->GetFieldID(*cls, name, "[B");
    if (fid == NULL)
        return false;

    jbyteArray jarray = (jbyteArray) env->GetObjectField(*object, fid);
    if (jarray == NULL)
        return false;

    jbyte* array = env->GetByteArrayElements(jarray, NULL);
    if (array == NULL)
    {
        env->DeleteLocalRef(jarray);
        return false;
    }

    memcpy(array, data, size*sizeof(BYTE));

    env->ReleaseByteArrayElements(jarray, array, 0);

    env->DeleteLocalRef(jarray);

    return true;
}

bool GetIntArrayField(JNIEnv * env, jclass* cls, jobject* object, char* name, int* data, int size)
{
    jfieldID fid = env->GetFieldID(*cls, name,   "[I");
    if (fid == NULL)
        return false;

    jintArray jarray = (jintArray) env->GetObjectField(*object, fid);
    if (jarray == NULL)
        return false;

    jint* array = env->GetIntArrayElements(jarray, NULL);
    if (array == NULL)
    {
        env->DeleteLocalRef(jarray);
        return false;
    }

    memcpy(data, array, size*sizeof(int));

    env->ReleaseIntArrayElements(jarray, array, 0);

    env->DeleteLocalRef(jarray);

    return true;
}

bool GetByteArrayField(JNIEnv * env, jclass* cls, jobject* object, char* name, BYTE* data, int size)
{
    jfieldID fid = env->GetFieldID(*cls, name, "[B");
    if (fid == NULL)
        return false;

    jbyteArray jarray = (jbyteArray) env->GetObjectField(*object, fid);
    if (jarray == NULL)
        return false;

    jbyte* array = env->GetByteArrayElements(jarray, NULL);
    if (array == NULL)
    {
        env->DeleteLocalRef(jarray);
        return false;
    }

    memcpy(data, array, size*sizeof(BYTE));

    env->ReleaseByteArrayElements(jarray, array, 0);

    env->DeleteLocalRef(jarray);

    return true;
}


jobject FillFeatures(JNIEnv * env, VFFEATURES &vf)
{
    jclass newExcCls = env->FindClass("VeriFinger/VeriFingerFeatures");
    if (newExcCls == NULL)
    {
        ThrowJavaClassNotFoundException(env);
        return NULL;
    }

    jmethodID constr = env->GetMethodID(newExcCls, "<init>", "()V");
    jobject result = (jobject) env->NewObject(newExcCls, constr);

    if (result == NULL)
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return NULL;
    }

    if (!SetIntArrayField(env, &newExcCls, &result, "m_x", vf.m.x, vf.m.count) ||
        !SetIntArrayField(env, &newExcCls, &result, "m_y", vf.m.y, vf.m.count) ||

        !SetByteArrayField(env, &newExcCls, &result, "m_d", vf.m.d, vf.m.count) ||
        !SetByteArrayField(env, &newExcCls, &result, "m_c", vf.m.c, vf.m.count) ||

        !SetIntArrayField(env, &newExcCls, &result, "sp_x", vf.sp.x, vf.sp.count) ||
        !SetIntArrayField(env, &newExcCls, &result, "sp_y", vf.sp.y, vf.sp.count) ||

        !SetByteArrayField(env, &newExcCls, &result, "sp_t", vf.sp.t, vf.sp.count) ||
        !SetByteArrayField(env, &newExcCls, &result, "sp_d", vf.sp.d, vf.sp.count))
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return NULL;
    }

    jfieldID fid_g          = env->GetFieldID(newExcCls, "g",           "B");
    jfieldID fid_m_count    = env->GetFieldID(newExcCls, "m_count",     "I");
    jfieldID fid_sp_count   = env->GetFieldID(newExcCls, "sp_count",    "I");
    jfieldID fid_bo_width   = env->GetFieldID(newExcCls, "bo_width",    "I");
    jfieldID fid_bo_height  = env->GetFieldID(newExcCls, "bo_height",   "I");

    jfieldID fid_bo_bits    = env->GetFieldID(newExcCls, "bo_bits",     "[[B");

    if (fid_g          == NULL ||
        fid_m_count    == NULL ||
        fid_sp_count   == NULL ||
        fid_bo_width   == NULL ||
        fid_bo_height  == NULL ||
        fid_bo_bits    == NULL)
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return NULL;
    }

    env->SetByteField(result, fid_g,         vf.g);
    env->SetIntField(result,  fid_m_count,   vf.m.count);
    env->SetIntField(result,  fid_sp_count,  vf.sp.count);
    env->SetIntField(result,  fid_bo_width,  vf.bo.width);
    env->SetIntField(result,  fid_bo_height, vf.bo.height);

    jobjectArray abo_bits = (jobjectArray) env->GetObjectField(result, fid_bo_bits);

    if (abo_bits == NULL)
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return NULL;
    }

    for (int i = 0; i < vf.bo.height; i++)
    {
        jobject obj = env->GetObjectArrayElement(abo_bits, i);
        if (obj == NULL)
        {
            env->DeleteLocalRef(abo_bits);

            env->DeleteLocalRef(newExcCls);
            ThrowJavaException(env);

            return NULL;
        }

        jbyte* el = env->GetByteArrayElements((jbyteArray)obj, NULL);

        if (el == NULL)
        {
            env->DeleteLocalRef(abo_bits);
            env->DeleteLocalRef(obj);

            env->DeleteLocalRef(newExcCls);
            ThrowJavaException(env);

            return NULL;
        }

        memcpy(el, &(vf.bo.bits[i][0]), sizeof(BYTE)*vf.bo.width);

        env->ReleaseByteArrayElements((jbyteArray)obj, el, 0);

        env->DeleteLocalRef(obj);
    }

    env->DeleteLocalRef(abo_bits);
    env->DeleteLocalRef(newExcCls);

    return result;
}

void GetFeatures(JNIEnv * env, jobject& jvf, VFFEATURES &vf)
{
    jclass newExcCls = env->FindClass("VeriFinger/VeriFingerFeatures");
    if (newExcCls == NULL)
    {
        ThrowJavaClassNotFoundException(env);
        return;
    }

    jfieldID fid_g          = env->GetFieldID(newExcCls, "g",           "B");
    jfieldID fid_m_count    = env->GetFieldID(newExcCls, "m_count",     "I");
    jfieldID fid_sp_count   = env->GetFieldID(newExcCls, "sp_count",    "I");
    jfieldID fid_bo_width   = env->GetFieldID(newExcCls, "bo_width",    "I");
    jfieldID fid_bo_height  = env->GetFieldID(newExcCls, "bo_height",   "I");

    jfieldID fid_bo_bits    = env->GetFieldID(newExcCls, "bo_bits",     "[[B");

    if (fid_g          == NULL ||
        fid_m_count    == NULL ||
        fid_sp_count   == NULL ||
        fid_bo_width   == NULL ||
        fid_bo_height  == NULL ||
        fid_bo_bits    == NULL)
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return;
    }

    vf.g         = env->GetByteField(jvf, fid_g);
    vf.m.count   = env->GetIntField(jvf,  fid_m_count);
    vf.sp.count  = env->GetIntField(jvf,  fid_sp_count);
    vf.bo.width  = env->GetIntField(jvf,  fid_bo_width);
    vf.bo.height = env->GetIntField(jvf,  fid_bo_height);

    if (!GetIntArrayField(env, &newExcCls, &jvf, "m_x", vf.m.x, vf.m.count) ||
        !GetIntArrayField(env, &newExcCls, &jvf, "m_y", vf.m.y, vf.m.count) ||

        !GetByteArrayField(env, &newExcCls, &jvf, "m_d", vf.m.d, vf.m.count) ||
        !GetByteArrayField(env, &newExcCls, &jvf, "m_c", vf.m.c, vf.m.count) ||

        !GetIntArrayField(env, &newExcCls, &jvf, "sp_x", vf.sp.x, vf.sp.count) ||
        !GetIntArrayField(env, &newExcCls, &jvf, "sp_y", vf.sp.y, vf.sp.count) ||

        !GetByteArrayField(env, &newExcCls, &jvf, "sp_t", vf.sp.t, vf.sp.count) ||
        !GetByteArrayField(env, &newExcCls, &jvf, "sp_d", vf.sp.d, vf.sp.count))
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return;
    }

    jobjectArray abo_bits = (jobjectArray) env->GetObjectField(jvf, fid_bo_bits);

    if (abo_bits == NULL)
    {
        env->DeleteLocalRef(newExcCls);
        ThrowJavaException(env);

        return;
    }

    for (int i = 0; i < vf.bo.height; i++)
    {
        jobject obj = env->GetObjectArrayElement(abo_bits, i);
        if (obj == NULL)
        {
            env->DeleteLocalRef(abo_bits);

            env->DeleteLocalRef(newExcCls);
            ThrowJavaException(env);

            return;
        }

        jbyte* el = env->GetByteArrayElements((jbyteArray)obj, NULL);

        if (el == NULL)
        {
            env->DeleteLocalRef(abo_bits);
            env->DeleteLocalRef(obj);

            env->DeleteLocalRef(newExcCls);
            ThrowJavaException(env);

            return;
        }

        memcpy(&(vf.bo.bits[i][0]), el, sizeof(BYTE)*vf.bo.width);

        env->ReleaseByteArrayElements((jbyteArray)obj, el, 0);

        env->DeleteLocalRef(obj);
    }

    env->DeleteLocalRef(abo_bits);
    env->DeleteLocalRef(newExcCls);
}

/*
 * Class:     VeriFinger_VeriFingerWrapper
 * Method:    VFDecompressFeatures
 * Signature: ([B)LVeriFinger/VeriFingerFeatures;
 */
JNIEXPORT jobject JNICALL Java_VeriFinger_VeriFingerWrapper_VFDecompressFeatures
  (JNIEnv * env, jclass cls, jbyteArray jfeatures)
{
    int res = VFE_OK;

    try
    {
        // get features
        jbyte* features = env->GetByteArrayElements(jfeatures, NULL);
        if (jfeatures == NULL || features == NULL)
            ThrowJavaIllegalArgumentException(env);
        else
        {
            VFFEATURES vf;
            memset(&vf, 0, sizeof(VFFEATURES));

            res = VFDecompressFeatures((const BYTE *)features, &vf);

            //if VFFailed(res)
            //    ThrowVeriFingerException(env, res);

            // release features
            env->ReleaseByteArrayElements(jfeatures, features, 0);

            // fill result object

            return FillFeatures(env, vf);
        }
    }
    catch(...)
    {
        ThrowJavaException(env);
    }

    return NULL;
}

/*
 * Class:     VeriFinger_VeriFingerWrapper
 * Method:    VFCompressFeatures
 * Signature: (LVeriFinger/VeriFingerFeatures;[B)I
 */
JNIEXPORT jint JNICALL Java_VeriFinger_VeriFingerWrapper_VFCompressFeatures
  (JNIEnv * env, jclass cls, jobject jvf, jbyteArray jfeatures)
{
    int res = VFE_OK;

    try
    {
        // get features
        jbyte* features = env->GetByteArrayElements(jfeatures, NULL);

        if (jfeatures == NULL || features == NULL)
            ThrowJavaIllegalArgumentException(env);
        else
        {
            VFFEATURES vf;
            memset(&vf, 0, sizeof(VFFEATURES));

            DWORD size = 0;

            GetFeatures(env, jvf, vf);

            size = VFCompressFeatures(&vf, (BYTE *)features);

            if VFFailed(res)
                ThrowVeriFingerException(env, res);

            // release features
            env->ReleaseByteArrayElements(jfeatures, features, 0);

            return size;
        }
    }
    catch(...)
    {
        ThrowJavaException(env);
    }

    return 0;
}
