/* $Id: gtk_job.c 9210 2006-09-19 07:43:25Z grazvydas $ */
#include <stdio.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <scanner.h>
#include <math.h> /* cos, sin */
#include <string.h>
#include <pthread.h>

#include "fp_engine.h"

#include "list.h"
#include "options.h"
#include "demo.h"
#include "thread.h"
#include "register_window.h"

#include "revision.h"
typedef struct gtk_global_struct {
	GtkWidget *window;
	GtkWidget *pixmap1;
	GtkWidget *pixmap2;
	GtkWidget *log;
	GtkWidget *manual_scan_menus[3];
	int       count;
	NFMinutia *minutiae;
} gtk_global_struct;

static gtk_global_struct gtk_items;

int log_printf (const char *format, ...);

static gboolean delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return FALSE;
}

static void destroy (GtkWidget *widget, gpointer data)
{
	if (scanner)
		scanner_close();
	gtk_main_quit ();
}

void set_title (void)
{
	char title[80], *name;
	NUInt ver_hi, ver_lo;
	int size;

	ver_hi = ver_lo = -1;

	size = 1 + VfeGetParameter (vf_extractor, VFEP_NAME, NULL);
	if (size < 1)
		return;
	name = (char *) malloc (size * sizeof(char));

	VfeGetParameter (vf_extractor, VFEP_NAME, name);
	VfeGetParameter (vf_extractor, VFEP_VERSION_HIGH, &ver_hi);
	VfeGetParameter (vf_extractor, VFEP_VERSION_LOW,  &ver_lo);
	snprintf (title, sizeof(title), "%s (%u.%u.%u.%u) Demo (rev: %s)", name,
		(ver_hi >> 16) & 0xFFFF, ver_hi & 0xFFFF,
		(ver_lo >> 16) & 0xFFFF, ver_lo & 0xFFFF,
		SRC_REVISION);
	free (name);
	gtk_window_set_title (GTK_WINDOW (gtk_items.window), title);
}

static void about_window ()
{
	GtkWidget *dialog;
	GtkWidget *text;

	log_printf ("About window\n");
	dialog = gtk_dialog_new_with_buttons ("About",
			GTK_WINDOW(gtk_items.window),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_STOCK_OK, GTK_RESPONSE_NONE, NULL);

	text = gtk_label_new (NULL);
	gtk_label_set_markup (GTK_LABEL (text),
			"<span size=\"xx-large\" foreground=\"red\">"
#ifdef VFDEMO
			"<b>VeriFinger</b></span>\n"
#else
#ifdef FCDEMO
			"<b>FingerCell</b></span>\n"
#else
#error select VFDEMO or FCDEMO
#endif
#endif
			"algorithm demonstration\n\n<span size=\"x-large\" "
			"foreground=\"blue\">Copyright (C) 2001-2006\n"
			"Neurotechnologija,\nAll rights reserved</span>"
			"\n\nE-mail: neurotec@neurotechnologija.com\n"
			"http://www.neurotechnologija.com");

	gtk_widget_set_size_request (text, 320, 260);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
			text);

	g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
			G_CALLBACK (gtk_widget_destroy),
			GTK_OBJECT (dialog));
	gtk_widget_show_all (dialog);
}

static char *get_image_from_file (gchar *selected_filename, int *W, int *H)
{
	GdkPixbuf	*pixbuf;
	GError		*error = NULL;
	char		*image;
	guchar		*pimage;
	int		i, n, x, y, w, h, s;

	*H = *W = 0;
	log_printf ("Image file: %s\n", selected_filename);

	pixbuf = gdk_pixbuf_new_from_file (selected_filename, &error);
	if (pixbuf == NULL)
	{
		log_printf ("load image failed: %s\n", error->message);
		return NULL;
	}

	gtk_image_set_from_pixbuf (GTK_IMAGE (gtk_items.pixmap1), pixbuf);

	w = gdk_pixbuf_get_width (pixbuf);
	h = gdk_pixbuf_get_height (pixbuf);
	n = gdk_pixbuf_get_rowstride (pixbuf);
	s = gdk_pixbuf_get_n_channels (pixbuf);
	pimage = gdk_pixbuf_get_pixels (pixbuf);
	if ((pimage == NULL) || (w <= 0) || (h <= 0) || (n <= 0))
		return NULL;

	image = (char *)malloc (w * h);
	if (image == NULL)
		return NULL;

	/* covert to raw 8bit format */
	i = 0;
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++)
		image [i++] = (	(int)(pimage[n * y + x * s + 0]) +
				(int)(pimage[n * y + x * s + 1]) +
				(int)(pimage[n * y + x * s + 2])  ) / 3;
	*H = h;
	*W = w;
	return image;
}

char *prompt_name (char *name, int mode)
{
	static char ret[80], *tmp;
	GtkWidget *dialog, *label;

	if (!(options.prompt || mode))
	{
		ret[0] = '\0';
		if (name)
			strncpy (ret, name, 80);
		return ret;
	}

	dialog = gtk_dialog_new_with_buttons ("Id input:",
				GTK_WINDOW (gtk_items.window),
				GTK_DIALOG_DESTROY_WITH_PARENT,
				GTK_STOCK_OK,
				GTK_RESPONSE_NONE,
				NULL);

	label = gtk_label_new ("Please enter finger name");
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);

	label = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY (label), name);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);

	g_signal_connect_swapped (GTK_OBJECT (dialog), "delete_event",
			G_CALLBACK (gtk_dialog_response),
			GTK_OBJECT (dialog));

	gtk_widget_show_all (dialog);

	gtk_dialog_run (GTK_DIALOG (dialog));

	tmp = (char *)gtk_entry_get_text (GTK_ENTRY (label));
	strncpy (ret, tmp, 80);
	gtk_widget_destroy (dialog);

	return ret;
}

void update_pixmap2 (char *image, int w, int h, int dpi)
{
	GdkPixbuf	*pixbuf;
	char		*image2;
	int		sw, sh, x, y, i, n, pos, ppos, dx, dy;
	int		circle[] = {0,3, 1,3, 2,2, 3,1, 3,0};
	float		scale = 1;

	scale = 500.0 / dpi;

	if ((w <= 0) || (h <= 0))
		return;

	sw = w * scale;
	sh = h * scale;

	image2 = (char *)malloc (3 * sw * sh);
	if (image2 == NULL)
		return;

	for (x = 0; x < sw; x++)
		for (y = 0; y < sh; y++)
		{
			image2 [3 * (y * sw + x) + 0] =
			image2 [3 * (y * sw + x) + 2] =
				image[(int)(((int)(y / scale)) * w + x / scale)];
			image2 [3 * (y * sw + x) + 1] = 0xFF;
		}

	for (i = 0; i < gtk_items.count; i++)
	{
		pos = (gtk_items.minutiae[i].Y * sw
			+ gtk_items.minutiae[i].X);
		for (n = 0; n < 5 * 2; n+=2)
		{
			ppos = 3 * (pos + circle [n] * sw + circle [n + 1]);
			image2 [ppos + 0] = 0xFF;
			image2 [ppos + 1] /= 2;
			image2 [ppos + 2] = 0;
			ppos = 3 * (pos + circle [n] * sw - circle [n + 1]);
			image2 [ppos + 0] = 0xFF;
			image2 [ppos + 1] /= 2;
			image2 [ppos + 2] = 0;
			ppos = 3 * (pos - circle [n] * sw + circle [n + 1]);
			image2 [ppos + 0] = 0xFF;
			image2 [ppos + 1] /= 2;
			image2 [ppos + 2] = 0;
			ppos = 3 * (pos - circle [n] * sw - circle [n + 1]);
			image2 [ppos + 0] = 0xFF;
			image2 [ppos + 1] /= 2;
			image2 [ppos + 2] = 0;
		}

		dx = cos(gtk_items.minutiae[i].Angle * M_PI / 128) * (+10);
		dy = sin(gtk_items.minutiae[i].Angle * M_PI / 128) * (+10);

		/* paint line */
		if (abs(dx) > abs(dy))
		{
			if (dx < 0)
			{
				for (x = dx; x < 0; x++)
				{
					y = x * dy / dx;
					ppos = 3 * (pos + y * sw + x);
					image2 [ppos + 0] = 0xFF;
					image2 [ppos + 1] /= 2;
					image2 [ppos + 2] = 0;
				}
			}
			else
			{
				for (x = 0; x < dx; x++)
				{
					y = x * dy / dx;
					ppos = 3 * (pos + y * sw + x);
					image2 [ppos + 0] = 0xFF;
					image2 [ppos + 1] /= 2;
					image2 [ppos + 2] = 0;
				}
			}
		}
		else
		{
			if (dy < 0)
			{
				for (y = dy; y < 0; y++)
				{
					x = y * dx / dy;
					ppos = 3 * (pos + y * sw + x);
					image2 [ppos + 0] = 0xFF;
					image2 [ppos + 1] /= 2;
					image2 [ppos + 2] = 0;
				}
			}
			else
			{
				for (y = 0; y < dy; y++)
				{
					x = y * dx / dy;
					ppos = 3 * (pos + y * sw + x);
					image2 [ppos + 0] = 0xFF;
					image2 [ppos + 1] /= 2;
					image2 [ppos + 2] = 0;
				}
			}
		}
	}

	pixbuf = gdk_pixbuf_new_from_data (image2,
			GDK_COLORSPACE_RGB, FALSE, 8, w * scale, h * scale,
			w * scale * 3, NULL, NULL);

	gtk_image_set_from_pixbuf (GTK_IMAGE (gtk_items.pixmap2), pixbuf);
}

void update_pixmap1 (char *image, int w, int h)
{
	GdkPixbuf	*pixbuf;
	char		*image2;
	int		i;

	if ((w <= 0) || (h <= 0))
		return;

	image2 = (char *)malloc (3 * w * h);
	if (image2 == NULL)
		return;

	for (i = 0; i < w * h; i++)
	{
		image2 [3 * i + 0] =
		image2 [3 * i + 1] =
		image2 [3 * i + 2] = image[i];
	}

	pixbuf = gdk_pixbuf_new_from_data (image2, GDK_COLORSPACE_RGB,
					FALSE, 8, w, h, w * 3, NULL, NULL);
	gtk_image_set_from_pixbuf (GTK_IMAGE (gtk_items.pixmap1), pixbuf);
}

void file_select_enroll (GtkWidget *widget, gpointer user_data)
{
	GtkWidget	*file_selector = (GtkWidget *)user_data;
	char		*selected_filename;
	char		*image;
	int		w, h;

	selected_filename = (char *)gtk_file_selection_get_filename
		(GTK_FILE_SELECTION (file_selector));

	image = get_image_from_file (selected_filename, &w, &h);
	if (image == NULL)
		return;

	if (enroll (image, w, h, options.resolution, selected_filename, 1))
		update_pixmap2 (image, w, h, options.resolution);
	free (image);
}

void file_select_generalize (GtkWidget *widget, gpointer user_data)
{
	GtkWidget	*file_selector = (GtkWidget *)user_data;
	char		*selected_filename;
	char		*image;
	int		w, h;

	selected_filename = (char *)gtk_file_selection_get_filename
		(GTK_FILE_SELECTION (file_selector));

	image = get_image_from_file (selected_filename, &w, &h);
	if (image == NULL)
		return;

	if (generalize (image, w, h, options.resolution, selected_filename, 1))
		update_pixmap2 (image, w, h, options.resolution);
	free (image);
}


void file_select_identify (GtkWidget *widget, gpointer user_data)
{
	GtkWidget	*file_selector = (GtkWidget *)user_data;
	char		*selected_filename;
	char		*image;
	int		w, h;

	selected_filename = (char *)gtk_file_selection_get_filename
		(GTK_FILE_SELECTION (file_selector));

	image = get_image_from_file (selected_filename, &w, &h);
	if (image == NULL)
		return;

	if (identify (image, w, h, options.resolution))
		update_pixmap2 (image, w, h, options.resolution);
	free (image);
}

static void file_select_menu (void *call_back)
{
	GtkWidget *filesel;

	filesel = gtk_file_selection_new ("Select images");
	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		"clicked", G_CALLBACK(call_back), filesel);
	g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION
		(filesel)->ok_button), "clicked", G_CALLBACK
		(gtk_widget_destroy), filesel);
	g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION
		(filesel)->cancel_button), "clicked", G_CALLBACK
		(gtk_widget_destroy), filesel);
	gtk_widget_show (filesel);
}

static void database_select_menu (void *call_back)
{
	GtkWidget *filesel;

	filesel = gtk_file_selection_new ("Select database file");
	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		"clicked", G_CALLBACK(call_back), filesel);
	g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION
		(filesel)->ok_button), "clicked", G_CALLBACK
		(gtk_widget_destroy), filesel);
	g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION
		(filesel)->cancel_button), "clicked", G_CALLBACK
		(gtk_widget_destroy), filesel);
	gtk_widget_show (filesel);
}

char *scan (int *w, int *h)
{
	char *image;
	struct timeval start_time;
	unsigned long time;

	if (scanner == NULL)
	{
		log_printf ("No scanner selected\n");
		return NULL;
	}

	timer_start (&start_time);
	image = scanner_read (w, h);
	time = timer_end (&start_time);

	log_printf ("Scanning time: %d.%06d\n", time / 1000000, time % 1000000);
	if (image == NULL)
	{
		log_printf ("Scan failed\n");
		return NULL;
	}
	return image;
}

void scanner_enroll (void)
{
	char		*image;
	int		w, h;

	image = scan (&w, &h);
	if (image == NULL)
		return;

	update_pixmap1 (image, w, h);
	if (enroll (image, w, h, scanner_dpi(), options.finger_id, 0))
		update_pixmap2 (image, w, h, scanner_dpi());
}

void scanner_generalize (void)
{
	char		*image;
	int		w, h;

	image = scan (&w, &h);
	if (image == NULL)
		return;

	update_pixmap1 (image, w, h);
	if (generalize (image, w, h, scanner_dpi(), options.finger_id, 0))
		update_pixmap2 (image, w, h, scanner_dpi());
}

void scanner_identify ()
{
	char		*image;
	int		w, h;

	image = scan (&w, &h);
	if (image == NULL)
		return;

	update_pixmap1 (image, w, h);
	if (identify (image, w, h, scanner_dpi()))
		update_pixmap2 (image, w, h, scanner_dpi());
}

void enable_continue_scan (int mode)
{
	static int state = 0;
	int i;

	switch (mode)
	{
		case 1:
		case 2:
		case 3:
			if (scanner == NULL)
			{
				log_printf ("No scanner selected\n");
				break;
			}
			con_mode = mode;
			if (!state)
			{
				log_printf ("continues scan enabled\n");
				pthread_mutex_unlock (&thr_mutex);
				state = 1;

				for (i=0; i < sizeof(gtk_items.manual_scan_menus)/sizeof(gtk_items.manual_scan_menus[0]); i++)
					gtk_widget_set_sensitive(gtk_items.manual_scan_menus[i], FALSE);
			}
			break;
		case 0:
			if (state)
			{
				pthread_mutex_lock (&thr_mutex);
				log_printf ("continues scan disabled\n");
				state = 0;

				for (i=0; i < sizeof(gtk_items.manual_scan_menus)/sizeof(gtk_items.manual_scan_menus[0]); i++)
					gtk_widget_set_sensitive(gtk_items.manual_scan_menus[i], TRUE);
			}
			break;
	}
}

static void menu_scanner_select_callback (GtkMenuItem *menuitem, gpointer user_data)
{
	int n = 0, i = (int)user_data;

	while (scanner_list[n++] && i--);
	if (i != -1)
		return;

	n--;
	scanner_close();
	scanner_set(scanner_list[n]);
	i = scanner_init();
	log_printf ("selected scanner: %s, dpi: %d, driver version %d, "
		"init returns: %d\n", scanner_name(), scanner_dpi(),
		scanner_version(), i);
}

void database_open (GtkWidget *widget, gpointer user_data)
{
	GtkWidget	*file_selector = (GtkWidget *)user_data;
	char		*selected_filename;
	void		*data;
	char		*name;
	FILE		*db;
	int		size, len;
	HNFRecord	record;
	NByte		g_value;

	selected_filename = (char *)gtk_file_selection_get_filename
		(GTK_FILE_SELECTION (file_selector));

	db = fopen (selected_filename, "r");
	if (db == NULL)
	{
		log_printf ("Can't open file:%s\n", selected_filename);
		return;
	}

	list_clear ();
	while (1)
	{
		if (fread (&len, sizeof(len), 1, db) != 1) break;
		if ((name = (char *)malloc(len + 1)) == NULL)
		{
			log_printf ("low mem\n");
			break;
		}
		if (fread (name, len, 1, db) != 1) break;
		name[len] = 0;

		if (fread (&size, sizeof(size), 1, db) != 1) break;
		if ((data = (char *)malloc(size)) == NULL)
		{
			log_printf ("low mem\n");
			break;
		}
		if (fread (data, size, 1, db) != 1) break;

		NFRecordCreateFromMemory(data, size, 0, NULL, &record);
		NFRecordGetG(record, &g_value);
		NFRecordFree(record);

		list_add (data, size, name, g_value);
	}
	fclose (db);
}

void database_save (GtkWidget *widget, gpointer user_data)
{
	GtkWidget	*file_selector = (GtkWidget *)user_data;
	char		*selected_filename;
	void		*data;
	char		*name;
	FILE		*db;
	int		size, len;

	selected_filename = (char *)gtk_file_selection_get_filename
		(GTK_FILE_SELECTION (file_selector));


	db = fopen (selected_filename, "w");
	if (db == NULL)
	{
		log_printf ("Can't create file:%s\n", selected_filename);
		return;
	}

	list_get_start ();
	while ((data = list_get_next(&name, &size)))
	{
		if (name)
			len = strlen (name);
		else
			len = 0;

		fwrite (&len, sizeof(len), 1, db);
		if (name)
			fwrite (name, len, 1, db);

		fwrite (&size, sizeof(size), 1, db);
		if (name)
			fwrite (data, size, 1, db);
	}
	fclose (db);
}

static void menu_callback (GtkMenuItem *menuitem, gpointer user_data)
{
	switch ((int)user_data)
	{
		case 1:
			file_select_menu (file_select_enroll);
			break;
		case 2:
			file_select_menu (file_select_identify);
			break;
		case 3:
			file_select_menu (file_select_generalize);
			break;
		case 4:
			scanner_enroll ();
			break;
		case 5:
			scanner_identify ();
			break;
		case 6:
			scanner_generalize ();
			break;
		case 7:
			if (scanner)
				scanner_close();
			gtk_main_quit ();
			break;
		case 8:
			enable_continue_scan (1);
			break;
		case 9:
			enable_continue_scan (2);
			break;
		case 10:
			enable_continue_scan (3);
			break;
		case 11:
			enable_continue_scan (0);
			break;
		case 12:
			options_window ();
			break;
		case 13:
			database_select_menu (database_open);
			break;
		case 14:
			database_select_menu (database_save);
			break;
		case 15:
			list_clear ();
			break;
		case 98:
			register_window ();
			break;
		case 99:
			about_window ();
			break;
		default:
			log_printf ("menu: %d\n", (int)user_data);
			break;
	}
}

void create_window (void)
{
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *menu_bar;
	GtkWidget *menu;
	GtkWidget *menu_shell;
	GtkWidget *menu_item;
	GtkWidget *seperator;
	GtkWidget *view_port;
	GtkWidget *log_scroll;
	int i;

	gtk_items.minutiae = NULL;

	/* create new window */
	gtk_items.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	g_signal_connect (G_OBJECT (gtk_items.window), "delete_event",
		G_CALLBACK (delete_event), NULL);

	g_signal_connect (G_OBJECT (gtk_items.window), "destroy",
		G_CALLBACK (destroy), NULL);

	gtk_container_set_border_width (GTK_CONTAINER (gtk_items.window), 2);

	/* create menu */
	menu = gtk_menu_new ();

	/* create menu_bar */
	menu_bar = gtk_menu_bar_new ();

	/* ---  --- */
	menu_item = gtk_menu_item_new_with_mnemonic ("_Enroll ...");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)1);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("_Identify ...");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)2);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("_Generalize ...");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)3);
	gtk_widget_show (menu_item);

	seperator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), seperator);
	gtk_widget_show (seperator);

	menu_item = gtk_menu_item_new_with_mnemonic ("E_nroll from scanner");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)4);
	gtk_widget_show (menu_item);
	gtk_items.manual_scan_menus[0] = menu_item;

	menu_item = gtk_menu_item_new_with_mnemonic ("I_dentify from scanner");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)5);
	gtk_widget_show (menu_item);
	gtk_items.manual_scan_menus[1] = menu_item;

	menu_item = gtk_menu_item_new_with_mnemonic ("Gene_ralize from scanner");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)6);
	gtk_widget_show (menu_item);
	gtk_items.manual_scan_menus[2] = menu_item;

	seperator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), seperator);
	gtk_widget_show (seperator);

	menu_item = gtk_menu_item_new_with_mnemonic ("E_xit");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)7);
	gtk_widget_show (menu_item);

	menu_shell = gtk_menu_item_new_with_mnemonic ("_Demonstration");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_shell), menu);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* --- / --- */

	/* ---  --- */
	menu = gtk_menu_new ();

	menu_item = gtk_menu_item_new_with_mnemonic ("with _Enroll");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)8);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("with _Identify");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)9);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("with _Generalize");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)10);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("_Stop");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)11);
	gtk_widget_show (menu_item);

	menu_shell = gtk_menu_item_new_with_mnemonic ("_Continuous scan");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_shell), menu);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* --- / --- */

	/* ---- Select scanner ------- */
	menu = gtk_menu_new ();

	for (i = 0; scanner_list[i]; i++)
	{
		menu_item = gtk_menu_item_new_with_mnemonic (scanner_list[i]->name);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
			G_CALLBACK (menu_scanner_select_callback), (void *)i);
		gtk_widget_show (menu_item);
	}

	menu_shell = gtk_menu_item_new_with_mnemonic ("_Select scanner");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_shell), menu);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* ------/ ----- */

	/* ---- Options ------- */
	menu_shell = gtk_menu_item_new_with_mnemonic ("_Options");
	g_signal_connect (G_OBJECT (menu_shell), "activate",
		G_CALLBACK (menu_callback), (void *)12);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* ------/ ----- */

	/* ---- Options ------- */
	menu = gtk_menu_new ();

	menu_item = gtk_menu_item_new_with_mnemonic ("_Open ...");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)13);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("_Save ...");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)14);
	gtk_widget_show (menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic ("_Clear");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
		G_CALLBACK (menu_callback), (void *)15);
	gtk_widget_show (menu_item);

	menu_shell = gtk_menu_item_new_with_mnemonic ("Da_tabase");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_shell), menu);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* ------/ ----- */

#if 0 /* ndef UNPROTECTED_DEMO */
	/* ---- Registration ------- */
	menu_shell = gtk_menu_item_new_with_mnemonic ("_Register");
	gtk_menu_item_set_right_justified (GTK_MENU_ITEM (menu_shell), TRUE);
	g_signal_connect (G_OBJECT (menu_shell), "activate",
		G_CALLBACK (menu_callback), (void *)98);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* ------/ ----- */
#endif

	/* ---- About ------- */
	menu_shell = gtk_menu_item_new_with_mnemonic ("_About");
	gtk_menu_item_set_right_justified (GTK_MENU_ITEM (menu_shell), TRUE);
	g_signal_connect (G_OBJECT (menu_shell), "activate",
		G_CALLBACK (menu_callback), (void *)99);
	gtk_widget_show (menu_shell);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_shell);
	/* ------/ ----- */

	/* vbox, vbox -> window */
	vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (gtk_items.window), vbox);
	gtk_widget_show (vbox);

	/* menu -> vbox */
	gtk_box_pack_start (GTK_BOX (vbox), menu_bar, FALSE, FALSE, 2);
	gtk_widget_show (menu_bar);

	/* hbox, hbox -> vbox */
	hbox = gtk_hbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (vbox), hbox);
	gtk_widget_show (hbox);

	/* pixmap1, -> vbox */
	gtk_items.pixmap1 = gtk_image_new ();
	view_port = gtk_viewport_new (NULL, NULL);
	gtk_container_add (GTK_CONTAINER (hbox), view_port);
	gtk_container_add (GTK_CONTAINER (view_port), gtk_items.pixmap1);
//	g_signal_connect (gtk_items.pixmap1, "expose_event",
//			G_CALLBACK (expose_cb), 1);
	gtk_widget_show (gtk_items.pixmap1);
	gtk_widget_show (view_port);
	gtk_widget_set_size_request (gtk_items.pixmap1, 320, 320);

	/* pixmap2, -> vbox */
	gtk_items.pixmap2 = gtk_image_new ();
	view_port = gtk_viewport_new (NULL, NULL);
	gtk_container_add (GTK_CONTAINER (hbox), view_port);
	gtk_container_add (GTK_CONTAINER (view_port), gtk_items.pixmap2);
//	g_signal_connect (gtk_items.pixmap2, "expose_event",
//			G_CALLBACK (expose_cb), 2);
	gtk_widget_show (gtk_items.pixmap2);
	gtk_widget_show (view_port);
	gtk_widget_set_size_request (gtk_items.pixmap2, 320, 320);

	/* log, -> hbox */
	log_scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_items.log = gtk_text_view_new ();
	gtk_container_add (GTK_CONTAINER (log_scroll), gtk_items.log);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW
			(log_scroll), GTK_POLICY_AUTOMATIC,
			GTK_POLICY_AUTOMATIC);
//	g_signal_connect (gtk_items.log, "expose_event",
//			G_CALLBACK (expose_cb), 0);
	gtk_widget_show (gtk_items.log);
	gtk_widget_show (log_scroll);
	gtk_container_add (GTK_CONTAINER (vbox), log_scroll);
	gtk_widget_set_size_request (log_scroll, 640, 230);
		/* set readonly & disable cursor */
	gtk_text_view_set_editable (GTK_TEXT_VIEW (gtk_items.log), FALSE);
	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (gtk_items.log), FALSE);
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (gtk_items.log), GTK_WRAP_WORD);

	gtk_widget_show  (gtk_items.window);
}

int log_printf (const char *format, ...)
{
	static int ident_level = 0;
	char buf[1024];
	va_list ap;
	GtkTextBuffer *buffer;
	GtkTextIter iter;
	GtkTextMark *mark;
	int ret;
	struct tm now;
	time_t now_time;

	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (gtk_items.log));


	/* decrease ident level on "<<<" */
	if (format[0] == '<' &&
	    format[1] == '<' &&
	    format[2] == '<')
		ident_level--;

	/* time */
	now_time = time (NULL);
	localtime_r (&now_time, &now);
	snprintf (buf, 1024, "%02d:%02d:%02d %*s", now.tm_hour,
			now.tm_min, now.tm_sec, ident_level*2, "");

	gtk_text_buffer_get_end_iter (buffer, &iter);
	gtk_text_buffer_insert (buffer, &iter, buf, -1);

	/* message */
	va_start (ap, format);
	ret = vsnprintf (buf, 1024, format, ap);
	gtk_text_buffer_get_end_iter (buffer, &iter);
	gtk_text_buffer_insert (buffer, &iter, buf, -1);

	/* scroll */
	gtk_text_buffer_get_end_iter (buffer, &iter);
	mark = gtk_text_buffer_create_mark (buffer, NULL, &iter, FALSE);
	gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (gtk_items.log),
					mark, 0, TRUE, 0, 0);

	gtk_widget_queue_draw (gtk_items.log);

	/* increase ident level on ">>>" */
	if (format[0] == '>' &&
	    format[1] == '>' &&
	    format[2] == '>')
		ident_level++;

	return ret;
}

void paint_minutia(HNFRecord record, NInt minutia_count)
{
	gtk_items.count = minutia_count;
	if (gtk_items.minutiae != NULL)
		free (gtk_items.minutiae);

	gtk_items.minutiae = (NFMinutia *) malloc(minutia_count * sizeof(NFMinutia));
	if (N_OK != NFRecordGetMinutiae(record, gtk_items.minutiae))
	{
		free (gtk_items.minutiae);
		gtk_items.minutiae = NULL;
		gtk_items.count = 0;
	}
}

