/* $Id: thread.c 9003 2006-09-15 11:45:29Z grazvydas $ */
#include <pthread.h>
#include <errno.h>
#include <scanner.h>

#include "demo.h"

static pthread_mutex_t scanner_muttex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t scanner_scanning = PTHREAD_MUTEX_INITIALIZER;

char *scanner_read (int *width, int *height)
{
	unsigned char * (*read) (int *width, int *height);
	char *ret = NULL;
	int i;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		read = scanner -> read;
	else
		read = NULL;
	pthread_mutex_unlock (&scanner_muttex);

	if (!read)
		return NULL;

	i = pthread_mutex_trylock (&scanner_scanning);
	if (i == EBUSY)
		return NULL;
	ret = read (width, height);
	pthread_mutex_unlock (&scanner_scanning);

	return ret;
}

void scanner_close (void)
{
	void (*close) (void);

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		close = scanner -> close;
	else
		close = NULL;
	scanner = NULL;
	pthread_mutex_unlock (&scanner_muttex);

	if (!close)
		return;

	pthread_mutex_lock (&scanner_scanning);
	close ();
	pthread_mutex_unlock (&scanner_scanning);
}

int scanner_init (void)
{
	int (*init) (void);
	int ret = -1;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
	{
		init = scanner -> init;
		scanner->max_bad_area = 80; /*   NOTE: MAKE THIS ACCESSIBLE FROM MENU   */
	}
	else
	{
		init = NULL;
	}
	pthread_mutex_unlock (&scanner_muttex);

	if (!init)
		return -1;

	pthread_mutex_lock (&scanner_scanning);
	ret = init ();
	pthread_mutex_unlock (&scanner_scanning);

	return ret;
}

int scanner_dpi (void)
{
	int dpi;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		dpi = scanner -> dpi;
	else
		dpi = 0;
	pthread_mutex_unlock (&scanner_muttex);

	return dpi;
}

char *scanner_name (void)
{
	char *name;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		name = scanner -> name;
	else
		name = NULL;
	pthread_mutex_unlock (&scanner_muttex);

	return name;
}

int scanner_version (void)
{
	int version;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		version = scanner -> version;
	else
		version = -1;
	pthread_mutex_unlock (&scanner_muttex);

	return version;
}

void scanner_set (struct scanner_info *item)
{
	pthread_mutex_lock (&scanner_muttex);
	scanner = item;
	pthread_mutex_unlock (&scanner_muttex);
}

