// FXDLL_LINUX must be defined for compilation under Linux

#ifndef HEADER_FX3DLL_H
#define HEADER_FX3DLL_H

#ifndef FXDLL_LINUX
	// win32
    #ifndef WINVER
    #include <windows.h>
    #endif

    #if WINVER<0x0400
    #error WINVER must be at least 0x0400
    #endif

  #ifdef _LIB
		#define FXDLL_EXPORTCALL
	#else
		#define FXDLL_EXPORTCALL __declspec(dllexport)
	#endif

	#define FXDLL_CALLMODE __stdcall
#else // linux
	
	#define FXDLL_EXPORTCALL 
	#define FXDLL_CALLMODE 
	#ifndef HWND
	#define HWND void*
	#endif
#endif // FXDLL_LINUX


#ifdef __cplusplus
extern "C" {
#endif

typedef int FX3error;
#define FX3_OK                        0
#define FX3_InsufficientMemory        1
#define FX3_TimeOut                   3
#define FX3_FileNotExisting           4
#define FX3_InvalidFingerprintFile    5
#define FX3_TooLowQuality             6
#define FX3_InvalidModelFile          7
#define FX3_ModelNotCreated           8
#define FX3_CannotWritetoDisk         9
#define FX3_CannotUpdateModel         10
#define FX3_InvalidMode               11
#define FX3_FingerprintNotAvailable   12
#define FX3_CannotProcessFingerprint  13
#define FX3_LowNumberOfMinutiae       14
#define FX3_ModelNotAvailable         15
#define FX3_Cancel                    16
#define FX3_InvalidSize               17

#define FX3_AcquisitionError          100
// Acquisition error specification (added to FX3_Acquisition Error)
  #define FX3SCAN_DllNotFound 10
  #define FX3SCAN_CorruptedDll 11
  #define FX3SCAN_InitializationFailure 12
  #define FX3SCAN_SettingFailure 13
  #define FX3SCAN_GrabbingFailure 14
  #define FX3SCAN_UnInitializationFailure 15
  #define FX3SCAN_ShowError 16
  #define FX3SCAN_TimerError 17
  #define FX3SCAN_InvalidParameters 18
  #define FX3SCAN_NoOnlineCapability 20
	#define FX3SCAN_InvalidFingerprintImage 21
	#define FX3SCAN_InvalidLatent 22
	#define FX3SCAN_InvalidLiveness 23
	#define FX3SCAN_FeatureNotAvailable 24
	#define FX3SCAN_CommunicationError 25
	#define FX3SCAN_FilePermissionError 26
	#define FX3SCAN_FileSystemError 27

// Operating modes
#define SPEED_DEFAULT 1
#define SPEED_FAST 2
#define SIZE_DEFAULT 1
#define SIZE_4K 2
#define SIZE_2K 3

// Acquisition Targets
#define TARGET_PC 1
#define TARGET_SCANNER 2

// Connections
#define CONNECTION_ON 1
#define CONNECTION_OFF 2

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_Init(void);
// Initialization; creates Fslot and Mslot
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   FX3_InsufficientMemory: insufficient memory for initialization

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_End(void);
// Releases memory and terminates
// RETURNED VALUE
//   FX3_OK: the function succeeded

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_Version(char* fx3,char* fx3scan,char* fx3enrdlg);
// Return the vesrion of fx3,fx3scan and fx3enrdlg dlls
// PARAMETERS
//         fx3: Input - pointer to the char array where the fx3.dll version will be copied.
//                      The array must be allocated by the caller (15 byte is the max size of the string returned)
//                      If the caller is not interested pass NULL          
//     fx3scan: Input - pointer to the char array where the fx3scan.dll version will be copied.
//                      The array must be allocated by the caller (15 byte is the max size of the string returned) 
//                      If the caller is not interested pass NULL
//   fx3enrdlg: Input - pointer to the char array where the fx3enrdlg.dll version will be copied.
//                      The array must be allocated by the caller (15 byte is the max size of the string returned) 
//                      If the caller is not interested pass NULL

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SetSessionKey(unsigned char* key);
// Set the session key used to encrypt/decrypt models;
//
//  FX3 uses the session key to encrypt/decrypt fingerprint models.
//  In order to create encrypted models, FX3_SetSessionKey must be called before:
//    - FX3_SaveModelToFile
//    - FX3_SaveModelToMemory
//  On the other hand, if a model has been encrypted by using a specific key,
//  the same key must loaded into the system (by using FX3_SetSessionKey)
//  before calling the following function to reload the model:  
//    - FX3_LoadModelFromFile
//    - FX3_LoadModelFromMemory
//    - FX3_LoadFingerprintFromModelOnFile
//    - FX3_LoadFingerprintFromModelOnMemory
//  The same session key remain valid until the FX3_End is called or another session
//    Key is loaded over the current one. 
//  Therefore it is RESPONSABILITY of the USER APPLICATION to STORE the KEYS used
//  to encrypt the models and RELOAD them into the system when required.   
//
// PARAMETERS
//   key: Input - pointer to the memory region (16 bytes) where a sesion key is stored
//        the session key may assume every possible value (2^128 distinct values)
// RETURNED VALUE
//   FX3_OK: the key has been setted

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_AcquireFingerprintOnLine(HWND hwnd,int x,int y,float* quality,int det_th,float timeout,float delay);
// Captures one fingerprint image on-line from the device into Fslot
//   Opens an Auto AGI dialog box which shows the live grabbing.
// PARAMETERS
//   hwnd: Input - handler of the window which will be the parent window of ACQDIAG.
//         NULL can be used for console applications or if an hwnd is not available.
//   x,y: Input - coordinates of the first corner (left,up) where ACQDIAG will be drawn. 
//        use negative valeus (i.e. -1,-1) to automatically center ACQDIAG in the screen.             
//   quality: Output - quality of the returned fingerprint image.
//   det_th: Input - detection presence threshold (0-255);
//           as soon as "detection" exceeds det_th a fingerprint image is captured and returned.      
//   t: Input - timeout (in seconds)
//   d: Input - delay (in seconds)

// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid fingerprint image has been copied in Fslot
//   FX3_Cancel: the user pressed the button 'Cancel'
//   FX3_TimeOut: the function returned because of the timeout
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_AcquireFingerprintOffLine(HWND hwnd,int x,int y,float* quality);
// Captures one fingerprint image off-line from the device into Fslot
//   Opens a Manual AGI dialog box which shows the live grabbing.
// PARAMETERS
//   hwnd: Input - handler of the window which will be the parent window of manual AGI dialog box.
//         NULL can be used for console applications or if an hwnd is not available.
//   x,y: Input - coordinates of the first corner (left,up) where the AGI will be drawn. 
//        use negative valeus (i.e. -1,-1) to automatically center the AGI on the screen. 
//   quality: Output - quality of the returned fingerprint image.
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid fingerprint image has been copied in Fslot
//   FX3_Cancel: the user pressed the button 'Cancel'
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_MuteAcquireFingerprintOnLine(float* quality,int det_th,float timeout,float delay);
// Behaves exactly like FX3_AcquireFingerprintOnLine, but no graphical display is performed

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_MuteAcquireFingerprintOnLineFast(unsigned char* imagep,int* Size_X,int* Size_Y,float* quality,int det_th,float timeout,float delay);
// INTERNAL USE ...
// Similarly to FX3_MuteAcquireFingerprintOnLine capture a fingerprint image as soon as "detection" exceeds det_th a fingerprint image
//   On the other hand the fingerprint image is not sent to the Fslot, but to the (user allcated) memory buffer imagep.
//   Image size is also returned in Size_X, Size_Y.
// The function FX3_LoadFingerprintFromMemory can be used to load the Fslot for matching.

#ifndef __PGRABCALLBACK
typedef int (*PGRABCALLBACK)(int presence100,int quality100,unsigned char *image,unsigned short ix,unsigned short iy);
#define __PGRABCALLBACK
#endif
FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_CallBackAcquireFingerprintOffLine(float* quality,PGRABCALLBACK cbFunc);
FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_CallBackAcquireFingerprintOnLine(float* quality,int det_th,float timeout,float delay,PGRABCALLBACK cbFunc);
// Behaves exactly like FX3_AcquireFingerprintOffLine and FX3_AcquireFingerprintOnLine respectively,
//   but a callback function is here provided for on-line updating


// ----------------- FXOEM Additional Commands ------------------------------
FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_Standby(void);
// Put the sensor in standy mode (power save)
// The sensor must be re-activated before calling acquisition/setting function  
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_Resume(void);
// Re-activate the sensor from standby mode.
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_GetConfig(unsigned char* LiveMode,unsigned char* InternalLight,unsigned char* RedLed,unsigned char* GreenLed);
// Read the sensor configuration: 
// If a NULL value is passed as a parameter pointer the corresponding setting is ignored
// PARAMETERS
//       LiveMode: Output - denotes the visual feedback mode during acquisition; use:
//                 1 for FULL_IMAGE
//                 2 for HALF_IMAGE
//                 3 for BLIND
//  InternalLight: Output - Internal light (red leds) activation ( 1:ON - 2:OFF )
//         RedLed: Output - Red led activation ( 1:ON - 2:OFF )
//       GreenLed: Output - Green Led activation ( 1:ON - 2:OFF )
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_SetConfig(unsigned char* LiveMode,unsigned char* InternalLight,unsigned char* RedLed,unsigned char* GreenLed);
// Set the sensor configuration: 
// If a NULL value is passed as a parameter pointer the corresponding setting is ignored
// PARAMETERS
//       LiveMode: Input - denotes the visual feedback mode during acquisition; use:
//                 1 for FULL_IMAGE (Dafault)
//                 2 for HALF_IMAGE
//                 3 for BLIND
//  InternalLight: Input - Internal light (red leds) activation ( 1:ON - 2:OFF )
//         RedLed: Input - Red led activation ( 1:ON - 2:OFF )
//       GreenLed: Input - Green Led activation ( 1:ON - 2:OFF )
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_SetAcquisitionTarget(unsigned char Target);
// Set the target of fingerprint acquisition. 
//   By defualt all the acquisition functions FX3_AcquireFingerprint... acquire a fingerprint image
//   on the PC (the default Target is the PC).
//   To perform Match On Board (only with scanners supporting MOB like FX3000) the acquisition
//   target must be the Scanner, so that the (last) fingerprint image is stored into the scanner
//   and not transferred to the PC. Even if the scanner is set as acquisition target, depending
//   on the LiveMode set with FX3_FX_SetConfig, the intermediate fingerprint images collcted during
//   the acquisition process can be still sent to the PC to provide a visual feedback to the user.
//   If one wants to avoid sending (intermediate) fingerprint images to the PC, he/she can
//   simply set LiveMode = BLIND.  
// PARAMETERS
//     Target: 1 TARGET_PC  (Default)
//						 2 TARGET_SCANNER
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_ForceScannerConnection(unsigned char Connection);
// By default each time an FX3_FX_??? or FX3_MOB_??? command have to be executed, a new communication
//   is established with the scanner and, at the end of the command, the communication channel is closed.
//   This is OK for most of the standard applications, while it could be not good for applications
//   using MOB capabilities of MOB scanners (like Fx3000). In fact, if the application needs to send
//   a sequence of FX3_MOB commands, opening and closing the communication channel at each single
//   command is not efficient. By using this command the application can force the communication channel
//   to remain open (CONNECTION_ON) until an explicit closing command is sent (CONNECTION_OFF).
//   The command also works for non-MOB scanners like Fx2000. 
// PARAMETERS
//     Connection: 1 CONNECTION_ON    
//								 2 CONNECTION_OFF  
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_SetFakeDetection(unsigned char LatentCheck,unsigned char LivenessCheck);
// Set the sensor fake detection mode: 
//   When fake detection is enabled the acquisition function may return an error (see the above FX3SCAN errors)
//   in case a fake finger has been detected
//       LatentCheck: Input - enables/disables check of latent fingerprint left on the sensor surface
//                    O Disabled
//                    1 Enabled
//     LivenessCheck: Input - enables/disables check of finger liveness (not yet implemented)
//										O Disabled
//                    1 Enabled
//       
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_EnterAutoDetection(void);
// Enter AutoDetection mode: this function must be called before FX3_FX_Presence
//  The internal illumination is activated and the two led turno off.
//  The sensor is ready to be polled by using FX3_FX_Presence
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_Presence(unsigned char* Presence);
// Polls the sensor to check the fingerprint presence
// Before using this function you must enter Autodetection mode (FX3_FX_EnterAutoDetection)
// At the end of AutoDetection you must call FX3_FX_ExitAutoDetection  
// The function takes about 150 ms before returning, even if practically
//   no CPU time is wasted since the PC CPU is internally released by the function during the sensor computations;   
//   therefore the PC can implement a strict polling routine without the risk of subtracting useful time
//   to other processes.  
// PARAMETERS
//     Presence: indicates the finger presence on the sensor
//       0: indicates minimum presence (no finger on the sensor)
//       255: indicates maxinum presence
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 


FXDLL_EXPORTCALL int FXDLL_CALLMODE FX3_FX_ExitAutoDetection(void);
// Exit AutoDetection mode: this function must be called to leave autodetection mode
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid fingerprint image has been copied in Fslot
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 
//   FX3_OK: the function succeeded
//   >= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_FX_SetActiveScanner(int ID);
// Assuming that more than one scanner is connected to the PC,
// sets as active scanner that identified by ID
//
// PARAMETERS
// ID - identifier of the scanner to be set as the active scanner
//  0: is the identifier for the default USB scanner; 
//  101..109: are valid identifiers for further USB scanners;
//  1: is the identifier for the default parallel port scanner; 
//  -1: is the identifier of the default scanner (the last one initialized through the control panel utility). 
// RETURNED VALUE
// 	FX3_OK: the function succeeded
//  	>= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_FX_GetActiveScanner(int* ID);
// Retrieves the ID of the active scanner
// PARAMETERS
// ID - identifier of the scanner to be set as the active scanner
//  0: is the identifier for the default USB scanner; 
//  101..109: are valid identifiers for further USB scanners;
//  1: is the identifier for the default parallel port scanner; 
//  -1: is the identifier of the default scanner (the last one initialized through the control panel utility). 
// RETURNED VALUE
// 	FX3_OK: the function succeeded
//  	>= FX3_AcquisitionError: an hardware/software problem has been encountered;  
//                            if > FX3_AcquisitionError the residual (value - FX3_AcquisitionError )
//                            defines the acuisition suberror occurred (see the above FX3SCAN errors) 

// ---------------------------------------------------------------------------

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadFingerprintFromFile(char* pathname,float* q);
// Loads one fingerprint image from file into Fslot
// PARAMETERS
//   pathname: Input - source file pathname
//   q: Output - quality of the loaded fingerprint image.
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid fingerprint image has been loaded in Fslot
//   FX3_FileNotExisting: the file denoted by pathname does not exist.
//   FX3_InvalidFingerprintFile: invalid file format.
//   FX3_TooLowQuality: the image quality is too low and it cannot be processed by FX3;
//                      no fingerprint has been loaded in Fslot.

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadFingerprintFromMemory(unsigned char* imagep,int X,int Y,float* q);
// Loads(copies) one fingerprint image from memory into Fslot
// PARAMETERS
//   imagep: Input - pointer to the memory region where an X by Y pixel image is stored by rows
//           from the top to the bottom
//        X: Input - image width
//        Y: Input - image height
//   q: Output - quality of the loaded fingerprint image.
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid fingerprint image has been loaded in Fslot
//   FX3_TooLowQuality: the image quality is too low and it cannot be processed by FX3;
//                      no fingerprint has been loaded in Fslot.
//   FX3_InvalidSize: the X or/and Y passed are not valid

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveFingerprintToFile(char* pathname);
// Saves the fingerprint image in Fslot to file
// Note that the image is saved in an internal format (e.g., 296x360 for FX scanners)
//   which is not interoperable with other third party recognition engines.
// Images saved by this function can be reloded into the Fslot by using the 
//   FX3_LoadImageFromFile function 
// PARAMETERS
//   pathname: Input - source file pathname
// RETURNED VALUE
//   FX3_OK: the function succeeded, the fingerprint image (in Fslot) has been stored
//   FX3_FingerprintNotAvailable: no fingerprints in Fslot
//   FX3_CannotWritetoDisk: the file specified by pathname cannot be written

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveFingerprintFullImageToFile(char* pathname);
// Saves the fingerprint image in Fslot to file
// The image is saved at FULL RESOLUTION (569 dpi, 296x560 pixels for FX scanners) and it is
//   suitable to be processed by third party recognition engines.
// Note that images "exported" by this function cannot be reloaded in FX3 SDK.
// PARAMETERS
//   pathname: Input - source file pathname
// RETURNED VALUE
//   FX3_OK: the function succeeded, the fingerprint image (in Fslot) has been stored
//   FX3_FingerprintNotAvailable: no fingerprints in Fslot
//   FX3_CannotWritetoDisk: the file specified by pathname cannot be written

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveFingerprintToMemory(unsigned char* imagep,int* X,int* Y);
// Saves(copies) the fingerprint image in Fslot to memory
//  The image is saved in internal format (see FX3_SaveFingerprintToFile) and can be
//    successively reloaded into Fslot by using FX3_LoadFingerprintFromMemory
// PARAMETERS
//   imagep: Input - pointer to the memory region where an X by Y pixel image will be stored by rows
//           from the top to the bottom;
//           the calling application is responsible of allocating (freeing) a X*Y bytes memory region
//           pointed by imagep
//           IMPORTANT: the user application can call this function with imagep = NULL
//                      to apriory known the values for X and Y and allocate the buffer
//                      imagep accordingly.     
//        X: Output - image width
//        Y: Output - image height
// RETURNED VALUE
//   FX3_OK: the function succeeded, the fingerprint image (in Fslot) has been stored
//   FX3_FingerprintNotAvailable: no fingerprints in Fslot

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveFingerprintFullImageToMemory(unsigned char* imagep,int* X,int* Y);
// Saves(copies) the fingerprint image in Fslot to memory
//  The image is saved(exported) at Full Resolution (see FX3_SaveFingerprintFullImageToFile) and cannot
//    be reloaded into Fslot.
// PARAMETERS
//   imagep: Input - pointer to the memory region where an X by Y pixel image will be stored by rows
//           from the top to the bottom;
//           the calling application is responsible of allocating (freeing) a X*Y bytes memory region
//           pointed by imagep
//           IMPORTANT: the user application can call this function with imagep = NULL
//                      to apriory known the values for X and Y and allocate the buffer
//                      imagep accordingly.     
//        X: Output - image width
//        Y: Output - image height
// RETURNED VALUE
//   FX3_OK: the function succeeded, the fingerprint image (in Fslot) has been stored
//   FX3_FingerprintNotAvailable: no fingerprints in Fslot

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadFingerprintFromModelOnFile(char* pathname);
// Loads a "virtual" fingerprint into Fslot from a model file
// PARAMETERS
//   pathname: Input - source file pathname
// RETURNED VALUE
//   FX3_OK: the function succeeded, a 1-model has been loaded into Fslot
//   FX3_FileNotExisting: the file denoted by pathname does not exist.
//   FX3_InvalidModelFile: invalid model format.

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadFingerprintFromModelOnMemory(unsigned char* modelp);
// Loads(copies) a "virtual" fingerprint into Fslot from a model file
// PARAMETERS
//   modelp: Input - pointer to the memory region where a valid model is stored
// RETURNED VALUE
//   FX3_OK: the function succeeded, a 1-model has been loaded into Fslot
//   FX3_InvalidModelFile: invalid model format.

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadModelFromFile(char* pathname);
// Loads a model from file into Mslot
// PARAMETERS
//   pathname: Input - source file pathname
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid model has been loaded in Mslot
//   FX3_FileNotExisting: the file denoted by pathname does not exist.
//   FX3_InvalidModelFile: invalid model format.

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_LoadModelFromMemory(unsigned char* modelp);
// Loads(copies) a model from memory into Mslot
// PARAMETERS
//   modelp: Input - pointer to the memory region where a valid model is stored
// RETURNED VALUE
//   FX3_OK: the function succeeded, a valid model has been loaded in Mslot
//   FX3_InvalidModelFile: invalid model format.

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveModelToFile(char* pathname,int mode,int* num);
// Writes the model from Fslot to file
// PARAMETERS
//   pathname: Input - target file pathname
//   mode: Input - use 0 for NEW, 1 for ADD
//   num: Output - number of fingerprints used to create the model
// RETURNED VALUE
//   FX3_OK: the function succeeded, the model from Fslot has been stored to disk
//   FX3_ModelNotCreated: no models in Fslot.
//   FX3_CannotWritetoDisk: I/O error, cannot save the file pathname to disk
//   FX3_InvalidModelFile: the model in Fslot is invalid
//   FX3_CannotUpdateModel: the model in Fslot cannot be added to pathname
//                          (no more than 3 fingerprint can be used).
//   FX3_InvalidMode: mode must be 0 or 1

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SaveModelToMemory(unsigned char* modelp,int mode,int* num,int* size);
// Writes the model from Fslot to file
// PARAMETERS
//   modelp: Input - pointer to the memory region where the model will be stored;
//           the calling application is responsible of allocating (freeing) a
//           MODELMAXMEMORY bytes memory region pointed by modelp
//   mode: Input - use 0 for NEW, 1 for ADD
//                 in case mode=ADD the existing model must be present in modelp before this call
//   num: Output - number of fingerprints used to create the model
//   size: Output - is the memory used to store the model
// RETURNED VALUE
//   FX3_OK: the function succeeded, the model from Fslot has been stored to memory
//   FX3_ModelNotCreated: no models in Fslot.
//   FX3_InvalidModelFile: the model in Fslot is invalid
//   FX3_CannotUpdateModel: the model in Fslot cannot be added to that in modelp
//                          (no more than 3 fingerprint can be used).
//   FX3_InvalidMode: mode must be 0 or 1

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_CreateModel(void);
// Creates a model in Fslot from the fingerprint image in Fslot
// RETURNED VALUE
//   FX3_OK: the function succeeded, a model has been stored in Fslot
//   FX3_LowNumberOfMinutiae: a low number of minutiae have been detected on the fingerprint in Fslot;
//                            anyway, a model has been stored in Fslot
//   FX3_FingerprintNotAvailable: no fingerprints in Fslot
//   FX3_CannotProcessFingerprint: the quality of the fingerprint in Fslot is too low;
//                                 a model cannot be created

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_Matching(float mins,float* similarity);
// matches the fingerprint image (or the 1-model) loaded in Fslot
//         with the (single or multiple) model loaded in Mslot.
// PARAMETERS
//   mins: Input - minimum similarity (for multiple-models)
//   similarity: Output - similarity resulting from the matching
// RETURNED VALUE
//   FX3_OK: the function succeeded
//   FX3_ModelNotAvailable: no model has been loaded in Mslot
//   FX3_FingerprintNotAvailable: neither fingerprint images nor 1-models in Fslot
//   FX3_CannotProcessFingerprint: the quality of the fingerprint image in Fslot is too low;
//                                 the matching cannot be performed

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_SetOperatingMode(int speed_mode, int size_mode);
// Set a given operating mode:
//  speed mode: SPEED_DEFAULT  (max. accuracy)
//              SPEED_FAST     (max. efficiency)
//   size_mode: SIZE_DEFAULT   (max. accuracy)
//              SIZE_4K        (modelsize <= 4K)
//              SIZE_2K        (modelsize <= 2K)
// RETURNED VALUE:
//   FX3_OK: the function succeeded
//   FX3_InvalidMode: Mode not supported

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_GetMaxModelSize(int* Size);
// Return the max model Size according to the Selected Operating Mode
// PARAMETERS
//   Size: Output - max size for a 1-model
// The Overall model Size is: 11+n*Size where n is the number of 1-models used.
// RETURNED VALUE
//   FX3_OK: the function succeeded

FXDLL_EXPORTCALL FX3error FXDLL_CALLMODE FX3_GetMaxModelSizeEx(int* Size,int X,int Y);
// Return the max model Size according to the Selected Operating Mode and the Image size (Full Image)
//   To get the image size (X, Y) use FX3_SaveFingerprintFullImageToMemory
// PARAMETERS
//   Size: Output - max size for a 1-model
// The Overall model Size is: 11+n*Size where n is the number of 1-models used.
// RETURNED VALUE
//   FX3_OK: the function succeeded

#ifdef __cplusplus
}
#endif

#endif // HEADER_FX3DLL_H