#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>

#define FXDLL_LINUX
#include <fx3dll.h>

#include "scanner.h"

//#define DEBUG

#ifdef DEBUG
#define logf(MSG, ...) printf(MSG, ##__VA_ARGS__)
#else
#define logf(MSG, ...) do { } while (0)
#endif

static unsigned char *scanner_read (int *wdth, int *hght);
static int scanner_init (void);
static void scanner_close (void);

struct scanner_info scanner_fx2k =
{
	.dpi 		569,
	.name 		"Fx2k/Fx3k",
	.version 	3,
	.close 		&scanner_close,
	.init 		&scanner_init,
	.read 		&scanner_read,
	.timeout	1000
};

static char *buf = NULL;
static int buf_size = 0;

static unsigned char *scanner_read(int *wdth, int *hght)
{
	int img_size;
	int fx_result;
	float fx_quality = 0.0;

	logf("scan...\n");
	fx_result = FX3_MuteAcquireFingerprintOnLine(&fx_quality, 70, (scanner_fx2k.timeout * 1.) / 1000, 0);
	logf("FX3_MuteAcquireFingerprintOnLine = %d\n", fx_result);
	if (fx_result != FX3_OK)
	{
		return NULL;
	}

	fx_result = FX3_SaveFingerprintFullImageToMemory(buf, wdth, hght);
	logf ("FX3_SaveFingerprintFullImageToMemory(size query) = %d\n", fx_result);
	if (fx_result != FX3_OK)
	{
		return NULL;
	}
	img_size = (*wdth) * (*hght);
	if (buf_size < img_size)
	{
		char *tbuf;
		tbuf = realloc(buf, img_size);
		if (tbuf == NULL)
		{
			logf("FX3 low mem: Unable to relocate buffer\n");
			return NULL;
		}

		buf = tbuf;
		buf_size = img_size;
	}

	fx_result = FX3_SaveFingerprintFullImageToMemory(buf, wdth, hght);
	logf("FX3_SaveFingerprintFullImageToMemory = %d\n", fx_result);
	if (fx_result != FX3_OK)
		return NULL;

	return buf;
}


static int scanner_init ()
{
	char v_fx3[16] = "";
	char v_fx3scan[16] = "";
	char v_fx3enrdlg[16] = "";
	int fx_result = FX3_OK;
	int id = 0;

	/* Fixme: set .dpi value for scanner_fx2k */

	fx_result = FX3_Init();
	logf("FX3_Init = %d\n", fx_result);
	if (fx_result != FX3_OK)
	{
		return 1;
	}

	/* Show driver library versions */
	FX3_Version(v_fx3, v_fx3scan, v_fx3enrdlg);
	logf("Fx scanner library Versions:\n"
	     "\tlibfx3: %s\n"
	     "\tlibfx3scan: %s\n"
	     "\tlibfx3enrdlg: %s\n", v_fx3, v_fx3scan, v_fx3enrdlg);


	fx_result = FX3_FX_GetActiveScanner(&id);
	logf("FX3_FX_GetActiveScanner = %d, id: %d\n", fx_result, id);

	/* Use this to change scanner used:
	 *         0: is the identifier for the default Fx2000 USB scanner;
	 *  101..109: are valid identifiers for further Fx2000 USB scanners;
	 *       200: is the identifier for the default Fx3000 USB scanner;
	 *  201..209: are valid identifiers for further Fx3000 USB scanners;
	 *         1: is the identifier for the default parallel port scanner;
	 *        -1: is the identifier of the default scanner (the last one
	 *  initialized through the control panel utility).
	 */
#if 0
	fx_result = FX3_FX_SetActiveScanner(200);
	logf("FX3_FX_SetActiveScanner = %d\n", fx_result);

	fx_result = FX3_FX_GetActiveScanner(&id);
	logf("FX3_FX_GetActiveScanner = %d, id: %d\n", fx_result, id);
#endif

	return (fx_result != FX3_OK);
}

static void scanner_close ()
{
	if (buf != NULL)
	{
		free(buf);
		buf = NULL;
		buf_size = 0;
	}

	FX3_End();
}
