#include <pthread.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <scanner.h>
#include <sys/time.h>
#include <glib.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_PNG
#define IMG_FTYPE ".png"
#include "png.c"
#else
#define IMG_FTYPE ".pgm"
int generate_img(char *name, char *img, int width, int height)
{
	FILE *f;

	f = fopen(name, "w");
	fprintf(f, "P5\n%d %d %d\n", width, height, 255);
	fwrite(img, width, height, f);

	fclose(f);

	return 0;
}
#endif


typedef struct gtk_global_struct {
	GtkWidget *window;
	GtkWidget *pixmap;
	GtkWidget *text;
	GtkObject *asc1;
	GtkObject *asc2;
} gtk_global_struct;

static gtk_global_struct	gtk_items;
static pthread_t		scan_thread;
static pthread_mutex_t		scanner_muttex = PTHREAD_MUTEX_INITIALIZER;

static struct scanner_info		*scanner_list[] = { SCANNERS NULL };
static struct scanner_info		*scanner;
static int working = 1;

static char *scanner_read (int *width, int *height)
{
	char *ret = NULL;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		ret =  scanner -> read (width, height);
	pthread_mutex_unlock (&scanner_muttex);

	return ret;
}

static void scanner_close (void)
{
	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		scanner -> close ();
	scanner = NULL;
	pthread_mutex_unlock (&scanner_muttex);
}

static void update_pixmap (char *image, int w, int h)
{
	GdkPixbuf	*pixbuf;
	char		*image0;
	static char	*image2;
	int		i;

	if ((w <= 0) || (h <= 0))
		return;

	image0 = image2;
	image2 = (char *)malloc (3 * w * h);
	if (image2 == NULL)
		return;

	if (image)
		for (i = 0; i < w * h; i++)
			image2 [3 * i + 0] =
			image2 [3 * i + 1] =
			image2 [3 * i + 2] = image[i];
	else
		memset (image2, -1, 3 * w * h);

	if (working)
	{
		pixbuf = gdk_pixbuf_new_from_data (image2, GDK_COLORSPACE_RGB,
					FALSE, 8, w, h, w * 3, NULL, NULL);
		gtk_image_set_from_pixbuf (GTK_IMAGE (gtk_items.pixmap), pixbuf);
		g_object_unref (pixbuf);
	}
	free (image0);
}

static void *continues_scan_loop (void *none)
{
	int w, h;
	char *image;
	char name[80];
	int skip = 0;
	int x, y;

	while (1)
	{
		image = scanner_read (&w, &h);
		if (image)
		{
			if (skip)
				continue;
			gdk_threads_enter ();
			update_pixmap (image, w, h);
			x = gtk_adjustment_get_value(
				GTK_ADJUSTMENT(gtk_items.asc1));
			y = gtk_adjustment_get_value(
				GTK_ADJUSTMENT(gtk_items.asc2));
			gdk_threads_leave ();

			snprintf(name, sizeof(name), "%s_%d_%d"IMG_FTYPE,
				gtk_entry_get_text(GTK_ENTRY(gtk_items.text)),
				x, y);
			generate_img(name, image, w, h);
			skip = 1;

			x++;
			if (x == 11)
			{
				x = 1;
				y++;
				if (y == 11)
					y = 1;
			}
			gdk_threads_enter ();
			gtk_adjustment_set_value(
				GTK_ADJUSTMENT(gtk_items.asc1), x);
			gtk_adjustment_set_value(
				GTK_ADJUSTMENT(gtk_items.asc2), y);
			gdk_threads_leave ();
		}
		else
		{
			skip = 0;
		}

	}

	return NULL;
}

static void continues_scan_init (void)
{
	pthread_create (&scan_thread, NULL, continues_scan_loop, NULL);
}

static gboolean delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return FALSE;
}

static void destroy (GtkWidget *widget, gpointer data)
{
	working = 0;
	scanner_close();
	gtk_main_quit ();
}

static void create_window (void)
{
	GtkWidget *vbox;
	GtkWidget *o;

	/* create new window */
	gtk_items.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	g_signal_connect (G_OBJECT (gtk_items.window), "delete_event",
		G_CALLBACK (delete_event), NULL);

	g_signal_connect (G_OBJECT (gtk_items.window), "destroy",
		G_CALLBACK (destroy), NULL);

	gtk_container_set_border_width (GTK_CONTAINER (gtk_items.window), 2);

	/* vbox */
	vbox = gtk_vbox_new (FALSE, 2);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (gtk_items.window), vbox);

	/* pixmap */
	gtk_items.pixmap = gtk_image_new ();
	gtk_widget_set_size_request (gtk_items.pixmap, 320, 320);
	gtk_widget_show (gtk_items.pixmap);
	gtk_container_add (GTK_CONTAINER (vbox), gtk_items.pixmap);

	/* input */
	gtk_items.text = gtk_entry_new ();
	gtk_widget_show (gtk_items.text);
	gtk_container_add (GTK_CONTAINER (vbox), gtk_items.text);

	/* finger */
	gtk_items.asc1 = gtk_adjustment_new(1, 1, 11, 1, 1, 1);
	o = gtk_hscrollbar_new(GTK_ADJUSTMENT(gtk_items.asc1));
	gtk_widget_show (o);
	gtk_container_add (GTK_CONTAINER (vbox), o);

	/* scann */
	gtk_items.asc2 = gtk_adjustment_new(1, 1, 11, 1, 1, 1);
	o = gtk_hscrollbar_new(GTK_ADJUSTMENT(gtk_items.asc2));
	gtk_widget_show (o);
	gtk_container_add (GTK_CONTAINER (vbox), o);

	gtk_widget_show  (gtk_items.window);
}

int main (int argc, char *argv[])
{
	int ret;

	scanner = scanner_list[0];
	if (scanner == NULL)
		return 1;

	g_thread_init (NULL);
	gdk_threads_init ();
	gdk_set_locale ();
	gtk_init (&argc, &argv);
	create_window ();

	ret = scanner -> init ();
	printf ("scanner %s initialization %s %d\n", scanner -> name, (ret == 0) ? "success" : "failed", ret);
	if (ret)
		return 2;

	continues_scan_init ();
	gtk_main ();

	if (scanner)
		scanner -> close ();

	return 0;
}
