#include <pthread.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <scanner.h>
#include <sys/time.h>
#include <glib.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

typedef struct gtk_global_struct {
	GtkWidget *window;
	GtkWidget *pixmap;
} gtk_global_struct;

static gtk_global_struct	gtk_items;
static pthread_t		scan_thread;
static pthread_mutex_t		scanner_muttex = PTHREAD_MUTEX_INITIALIZER;

static struct scanner_info		*scanner_list[] = { SCANNERS NULL };
static struct scanner_info		*scanner;
static int working = 1;

static char *scanner_read (int *width, int *height)
{
	char *ret = NULL;

	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		ret =  scanner -> read (width, height);
	pthread_mutex_unlock (&scanner_muttex);

	return ret;
}

static void scanner_close (void)
{
	pthread_mutex_lock (&scanner_muttex);
	if (scanner)
		scanner -> close ();
	scanner = NULL;
	pthread_mutex_unlock (&scanner_muttex);
}

static void update_pixmap (char *image, int w, int h)
{
	GdkPixbuf	*pixbuf;
	char		*image0;
	static char	*image2;
	int		i;

	if ((w <= 0) || (h <= 0))
		return;

	image0 = image2;
	image2 = (char *)malloc (3 * w * h);
	if (image2 == NULL)
		return;

	if (image)
		for (i = 0; i < w * h; i++)
			image2 [3 * i + 0] =
			image2 [3 * i + 1] =
			image2 [3 * i + 2] = image[i];
	else
		memset (image2, 0x80, 3 * w * h);

	if (working)
	{
		pixbuf = gdk_pixbuf_new_from_data (image2, GDK_COLORSPACE_RGB,
					FALSE, 8, w, h, w * 3, NULL, NULL);
		gtk_image_set_from_pixbuf (GTK_IMAGE (gtk_items.pixmap), pixbuf);
		g_object_unref (pixbuf);
	}
	free (image0);
}

static void *continues_scan_loop (void *none)
{
	int w, h;
	char *image;

	while (1)
	{
		image = scanner_read (&w, &h);
		gdk_threads_enter ();
		update_pixmap (image, w, h);
		gdk_threads_leave ();
	}

	return NULL;
}

static void continues_scan_init (void)
{
	pthread_create (&scan_thread, NULL, continues_scan_loop, NULL);
}

static gboolean delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return FALSE;
}

static void destroy (GtkWidget *widget, gpointer data)
{
	working = 0;
	scanner_close();
	gtk_main_quit ();
}

static void create_window (void)
{
	/* create new window */
	gtk_items.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	g_signal_connect (G_OBJECT (gtk_items.window), "delete_event",
		G_CALLBACK (delete_event), NULL);

	g_signal_connect (G_OBJECT (gtk_items.window), "destroy",
		G_CALLBACK (destroy), NULL);

	gtk_container_set_border_width (GTK_CONTAINER (gtk_items.window), 2);

	/* pixmap */
	gtk_items.pixmap = gtk_image_new ();
	gtk_container_add (GTK_CONTAINER (gtk_items.window), gtk_items.pixmap);
	gtk_widget_show (gtk_items.pixmap);
	gtk_widget_set_size_request (gtk_items.pixmap, 320, 320);

	gtk_widget_show  (gtk_items.window);
}

int main (int argc, char *argv[])
{
	int ret;

	scanner = scanner_list[0];
	if (scanner == NULL)
		return 1;

	g_thread_init (NULL);
	gdk_threads_init ();
	gdk_set_locale ();
	gtk_init (&argc, &argv);
	create_window ();

	printf("\n\nGTK+ scanner demo. Copyright (C) 1998-2006 Neurotechnologija\n\n");

	ret = scanner -> init ();
	printf ("scanner %s initialization: %d(%s)\n", scanner -> name, ret, (ret == 0) ? "success" : "failed");
	if (ret)
		return 2;

	continues_scan_init ();
	gtk_main ();

	if (scanner)
		scanner -> close ();

	return 0;
}
